(* Author: Dmitriy Traytel *)

header {* Deciding Equivalence of $\Pi$-Extended Regular Expressions *}

(*<*)
theory Equivalence_Checking
imports Derivatives "~~/src/HOL/Library/While_Combinator"
  "~~/src/HOL/Library/Product_Lexorder" "~~/src/HOL/Library/RBT_Set"
begin

hide_const RBT.map
(*>*)

subsection {* Bisimulation between languages and regular expressions *}

type_synonym 'a rexp_pair = "'a list list * ('a list rexp * 'a list rexp)"
type_synonym 'a rexp_pairs = "'a rexp_pair list"

context alphabet
begin

context
fixes n :: nat
begin

coinductive bisimilar :: "'a lang \<Rightarrow> 'a lang \<Rightarrow> bool" where
"K \<subseteq> lists (\<Sigma> n) \<Longrightarrow> L \<subseteq> lists (\<Sigma> n)
 \<Longrightarrow> ([] \<in> K \<longleftrightarrow> [] \<in> L)
 \<Longrightarrow> (\<And>x. x \<in> \<Sigma> n \<Longrightarrow> bisimilar (lQuot x K) (lQuot x L))
 \<Longrightarrow> bisimilar K L"

lemma equal_if_bisimilar:
assumes "K \<subseteq> lists (\<Sigma> n)" "L \<subseteq> lists (\<Sigma> n)" "bisimilar K L" shows "K = L"
proof (rule set_eqI)
  fix w
  from assms show "w \<in> K \<longleftrightarrow> w \<in> L"
  proof (induction w arbitrary: K L)
    case Nil thus ?case by (auto elim: bisimilar.cases)
  next
    case (Cons a w K L)
    show ?case
    proof cases
      assume "a \<in> \<Sigma> n"
      with `bisimilar K L` have "bisimilar (lQuot a K) (lQuot a L)"
        by (auto elim!: bisimilar.cases)
      then have "w \<in> lQuot a K \<longleftrightarrow> w \<in> lQuot a L"
        by (intro Cons.IH) (auto elim!: bisimilar.cases)
      thus ?case by (auto simp: lQuot_def)
    next
      assume "a \<notin> \<Sigma> n"
      thus ?case using Cons.prems by fastforce
    qed
  qed
qed

lemma language_coinduct:
fixes R (infixl "\<sim>" 50)
assumes "\<And>K L. K \<sim> L \<Longrightarrow> K \<subseteq> lists (\<Sigma> n) \<and> L \<subseteq> lists (\<Sigma> n)"
assumes "K \<sim> L"
assumes "\<And>K L. K \<sim> L \<Longrightarrow> ([] \<in> K \<longleftrightarrow> [] \<in> L)"
assumes "\<And>K L x. K \<sim> L \<Longrightarrow> x : \<Sigma> n \<Longrightarrow> lQuot x K \<sim> lQuot x L"
shows "K = L"
apply (rule equal_if_bisimilar)
apply (rule conjunct1[OF assms(1)[OF assms(2)]])
apply (rule conjunct2[OF assms(1)[OF assms(2)]])
apply (rule bisimilar.coinduct[of R, OF `K \<sim> L`])
apply (auto simp: assms)
done

end

end

context embed
begin

definition is_bisimulation where
"is_bisimulation n X =
  (\<forall>(r,s) \<in> X. wf n r \<and> wf n s \<and> (final r \<longleftrightarrow> final s) \<and>
  (\<forall>a\<in>\<Sigma> n. (\<guillemotleft>lderiv a r\<guillemotright>, \<guillemotleft>lderiv a s\<guillemotright>) \<in> X))"

lemma bisim_lang_eq:
fixes r s :: "'a rexp"
assumes bisim: "is_bisimulation n X"
assumes "(r, s) \<in> X"
shows "lang n r = lang n s"
proof -
  let ?R = "\<lambda>K L. (\<exists>(r,s) \<in> X. K = lang n r \<and> L = lang n s)"
  show ?thesis
  proof (rule language_coinduct[where R="?R"])
    from `(r, s) \<in> X` bisim show "?R (lang n r) (lang n s)"
      by (auto split: prod.splits simp:  is_bisimulation_def)
  next
    fix K L assume "?R K L"
    then obtain r s where rs: "(r, s) \<in> X"
      and KL: "K = lang n r" "L = lang n s" by auto
    with bisim have "final r \<longleftrightarrow> final s" and wfr: "wf n r" and wfs: "wf n s"
      by (auto simp: is_bisimulation_def)
    thus "[] \<in> K \<longleftrightarrow> [] \<in> L"
      by (auto simp: lang_final[of r n] lang_final[of s n] KL)
  txt{* next case, but shared context *}
    from bisim rs KL lang_subset_lists
    show "K \<subseteq> lists (\<Sigma> n) \<and> L \<subseteq> lists (\<Sigma> n)"
      unfolding is_bisimulation_def by fastforce
  txt{* next case, but shared context *}
    fix a assume *: "a \<in> \<Sigma> n"
    with rs bisim have witness: "(\<guillemotleft>lderiv a r\<guillemotright>, \<guillemotleft>lderiv a s\<guillemotright>) \<in> X"
      by (fastforce simp: is_bisimulation_def)
    show "?R (lQuot a K) (lQuot a L)"
      using KL ACI_norm_lang lang_lderiv[OF wfr *] lang_lderiv[OF wfs *]
      by (blast intro!: bexI[OF _ witness])
  qed
qed

lemma lderivs_lang_eq:
fixes r s :: "'a rexp"
assumes "wf n r" "wf n s"
shows "(\<forall>(r, s) \<in> lderivs_set n \<guillemotleft>r\<guillemotright> \<guillemotleft>s\<guillemotright>. final r = final s) = (lang n r = lang n s)" (is "?L = ?R")
proof
  assume ?L
  hence "\<forall>(r, s) \<in> lderivs_set n \<guillemotleft>r\<guillemotright> \<guillemotleft>s\<guillemotright>. wf n r \<and> wf n s  \<and> (final r \<longleftrightarrow> final s)"
    using assms by (auto simp add: ACI_norm_lderivs)
  moreover
  { fix r' s' w assume "(r', s') \<in> lderivs_set n r s" and *: "w \<in> \<Sigma> n"
    then obtain ws where ws: "wf_word n ws" "r' = \<guillemotleft>lderivs ws r\<guillemotright>"  "s' = \<guillemotleft>lderivs ws s\<guillemotright>" by auto
    with * have "(\<guillemotleft>lderiv w r'\<guillemotright>, \<guillemotleft>lderiv w s'\<guillemotright>) = (\<guillemotleft>lderivs (ws @ [w]) r\<guillemotright>, \<guillemotleft>lderivs (ws @ [w]) s\<guillemotright>)"
        by (auto simp: ACI_norm_lderiv)
    hence "(\<guillemotleft>lderiv w r'\<guillemotright>, \<guillemotleft>lderiv w s'\<guillemotright>) \<in> lderivs_set n r s"
      using * ws(1) by (auto intro!: imageI exI[of _ "ws @ [w]"])
  }
  ultimately have "is_bisimulation n (lderivs_set n \<guillemotleft>r\<guillemotright> \<guillemotleft>s\<guillemotright>)"
    unfolding is_bisimulation_def by (auto simp: ACI_norm_lderivs)
  hence "lang n \<guillemotleft>r\<guillemotright> = lang n \<guillemotleft>s\<guillemotright>" by (intro bisim_lang_eq) (auto intro: exI[of _ "[]"])
  thus ?R by (rule box_equals[OF _ ACI_norm_lang ACI_norm_lang])
next
  assume ?R thus ?L using assms lang_lderivs lang_final by (auto simp: ACI_norm_lderivs) metis+
qed

end



subsection {* Different normalization function *}

locale normalizer = embed \<Sigma> project embed
  for \<Sigma> :: "nat \<Rightarrow> 'a :: linorder set"
  and project :: "'a \<Rightarrow> 'a"
  and embed :: "'a \<Rightarrow> 'a list" +
fixes norm :: "'a :: linorder rexp \<Rightarrow> 'c"
and   nlang :: "nat \<Rightarrow> 'c \<Rightarrow> 'a list set"
assumes lang_norm: "wf n r \<Longrightarrow> nlang n (norm r) = lang n r"
begin

abbreviation "nfinal n r \<equiv> ([] \<in> nlang n r)"

lemma nfinal_final: "wf n r \<Longrightarrow> nfinal n (norm r) = final r"
  using lang_final lang_norm by blast

definition "norms \<equiv> (%(r,s). (norm r, norm s))"

lemma finite_norm: "finite {norm \<guillemotleft>lderivs xs r\<guillemotright> | xs . True}"
  by (rule finite_surj[OF finite_lderivs, of _ norm]) auto

lemma finite_norm_lderivs: "finite (norms ` (lderivs_set n r s))"
  by (intro finite_subset[OF _ finite_cartesian_product[OF finite_norm finite_norm]]) (auto simp: norms_def)

definition is_nbisimulation where
"is_nbisimulation n X =
  (\<forall>(r,s) \<in> X. wf n r \<and> wf n s \<and> (final r \<longleftrightarrow> final s) \<and>
  (\<forall>a\<in>\<Sigma> n. (norm \<guillemotleft>lderiv a r\<guillemotright>, norm \<guillemotleft>lderiv a s\<guillemotright>) \<in> norms ` X))"

lemma nbisim_lang_eq:
  fixes r s :: "'a rexp"
  assumes nbisim: "is_nbisimulation n X"
  assumes "(r, s) \<in> X"
  shows "lang n r = lang n s"
proof -
  let ?R = "\<lambda>K L. (\<exists>(r,s) \<in> norms ` X. K = nlang n r \<and> L = nlang n s)"
  show ?thesis
  proof (rule language_coinduct[where R="?R"])
    from `(r, s) \<in> X` nbisim show "?R (lang n r) (lang n s)"
      by (auto split: prod.splits simp: lang_norm norms_def is_nbisimulation_def)
  next
    fix K L assume "?R K L"
    then obtain r s where rs: "(r, s) \<in> X" 
      and KL: "K = nlang n (norm r)" "L = nlang n (norm s)" by (auto simp: norms_def)
    with nbisim have "final r \<longleftrightarrow> final s" and wfr: "wf n r" and wfs: "wf n s"
      by (auto simp: is_nbisimulation_def)
    thus "[] \<in> K \<longleftrightarrow> [] \<in> L"
      by (auto simp: lang_norm[OF wfr] lang_norm[OF wfs] lang_final[of r n] lang_final[of s n] KL)
  txt{* next case, but shared context *}
    from nbisim rs KL lang_subset_lists
    show "K \<subseteq> lists (\<Sigma> n) \<and> L \<subseteq> lists (\<Sigma> n)"
      unfolding is_nbisimulation_def lang_norm[OF wfr] lang_norm[OF wfs] by fastforce
  txt{* next case, but shared context *}
    fix a assume *: "a \<in> \<Sigma> n"
    with rs nbisim have witness: "(norm \<guillemotleft>lderiv a r\<guillemotright>, norm \<guillemotleft>lderiv a s\<guillemotright>) \<in> norms ` X"
      by (fastforce simp: is_nbisimulation_def)
    show "?R (lQuot a K) (lQuot a L)"
      using KL[unfolded lang_norm[OF wfr] lang_norm[OF wfs]]
        trans[OF lang_norm[OF iffD2[OF ACI_norm_wf, OF wf_lderiv[OF wfr]]] ACI_norm_lang]
        trans[OF lang_norm[OF iffD2[OF ACI_norm_wf, OF wf_lderiv[OF wfs]]] ACI_norm_lang]
        lang_lderiv[OF wfr *] lang_lderiv[OF wfs *]
      by (blast intro!: bexI[OF _ witness])
  qed
qed

lemma norm_lderivs_lang_eq:
fixes r s :: "'a rexp"
assumes "wf n r" "wf n s"
shows "(\<forall>(r, s) \<in> norms ` lderivs_set n \<guillemotleft>r\<guillemotright> \<guillemotleft>s\<guillemotright>. nfinal n r = nfinal n s) = (lang n r = lang n s)"
  by (rule trans[OF _ lderivs_lang_eq[OF assms]]) (fastforce simp: norms_def assms nfinal_final)

end


text {* Closure computation *}

primrec remdups' where
  "remdups' f [] = []"
| "remdups' f (x # xs) =
    (case List.find (\<lambda>y. f x = f y) xs of None \<Rightarrow> x # remdups' f xs | _ \<Rightarrow> remdups' f xs)"

lemma map_remdups'[simp]: "map f (remdups' f xs) = remdups (map f xs)"
  by (induct xs) (auto split: option.splits simp add: find_Some_iff find_None_iff)

lemma remdups'_map[simp]: "remdups' f (map g xs) = map g (remdups' (f o g) xs)"
  by (induct xs) (auto split: option.splits simp add: find_None_iff,
                auto simp: find_Some_iff elim: imageI[OF nth_mem])

lemma map_apfst_remdups':
  "map (f o fst) (remdups' snd xs) = map fst (remdups' snd (map (apfst f) xs))"
  by (auto simp: comp_def)

lemma set_remdups'[simp]: "f ` set (remdups' f xs) = f ` set xs"
  by (induct xs) (auto split: option.splits simp add: find_Some_iff)

lemma subset_remdups': "set (remdups' f xs) \<subseteq> set xs"
  by (induct xs) (auto split: option.splits)

lemma find_append[simp]:
  "List.find P (xs @ ys) = None = (List.find P xs = None \<and> List.find P ys = None)"
  by (induct xs) auto

lemma subset_remdups'_append: "set (remdups' f (xs @ ys)) \<subseteq> set (remdups' f xs) \<union> set (remdups' f ys)"
  by (induct xs arbitrary: ys) (auto split: option.splits)

lemmas mp_remdups' = set_mp[OF subset_remdups']
lemmas mp_remdups'_append = set_mp[OF subset_remdups'_append]

lemma inj_on_set_remdups'[simp]: "inj_on f (set (remdups' f xs))"
  by (induct xs) (auto split: option.splits simp add: find_None_iff dest!: mp_remdups')

lemma distinct_remdups'[simp]: "distinct (map f (remdups' f xs))"
  by (induct xs) (auto split: option.splits simp: find_None_iff)

lemma distinct_remdups'_strong: "(\<forall>x\<in>set xs. \<forall>y\<in>set xs. g x = g y \<longrightarrow> f x = f y) \<Longrightarrow>
  distinct (map g (remdups' f xs))"
proof (induct xs)
  case (Cons x xs) thus ?case
    by (auto split: option.splits) (fastforce simp: find_None_iff dest!: mp_remdups')
qed simp

lemma set_remdups'_strong: "(\<forall>x\<in>set xs. \<forall>y\<in>set xs. g x = g y \<longrightarrow> f x = f y) \<Longrightarrow>
  f ` set (remdups' g xs) = f ` set xs"
proof (induct xs)
  case (Cons x xs) thus ?case
    by (clarsimp split: option.splits simp add: find_Some_iff)
       (intro insert_absorb[symmetric] image_eqI[OF _ nth_mem, of _ f xs], auto)
qed simp

fun test where "test (ws, _, _) = (case ws of [] \<Rightarrow>  False | (w,p,q)#_ \<Rightarrow> final p = final q)"
fun test' where "test' (ws, _) = (case ws of [] \<Rightarrow>  False | (p,q)#_ \<Rightarrow> final p = final q)"

locale equivalence_checker =
fixes \<sigma> :: "nat \<Rightarrow> 'a :: linorder list"
and \<pi> :: "'a \<Rightarrow> 'a"
and \<epsilon> :: "'a \<Rightarrow> 'a list"
and norm :: "'a rexp \<Rightarrow> 'c"
and   nlang :: "nat \<Rightarrow> 'c \<Rightarrow> 'a list set"
assumes norm: "normalizer (set \<circ> \<sigma>) \<pi> \<epsilon> norm nlang"

sublocale equivalence_checker \<subseteq> normalizer "set \<circ> \<sigma>" \<pi> \<epsilon>
  by (rule norm)

context equivalence_checker
begin

fun step where "step n (ws, ps, N) =
    (let
      (w, r, s) = hd ws;
      ps' = (r, s) # ps;
      succs = map (\<lambda>a.
        let
          r' = \<guillemotleft>lderiv a r\<guillemotright>;
          s' = \<guillemotleft>lderiv a s\<guillemotright>
        in ((a # w, r', s'), (norm r', norm s'))) (\<sigma> n);
      new = remdups' snd (filter (\<lambda>(_, rs). rs \<notin> N) succs);
      ws' = tl ws @ map fst new;
      N' = set (map snd new) \<union> N
    in (ws', ps', N'))"

fun step' where "step' n (ws, N) =
    (let
      (r, s) = hd ws;
      succs = map (\<lambda>a.
        let
          r' = \<guillemotleft>lderiv a r\<guillemotright>;
          s' = \<guillemotleft>lderiv a s\<guillemotright>
        in ((r', s'), (norm r', norm s'))) (\<sigma> n);
      new = remdups' snd (filter (\<lambda>(_, rs). rs \<notin> N) succs)
    in (tl ws @ map fst new, set (map snd new) \<union> N))"

lemma step_unfold: "step n (w # ws, ps, N) = (ws', ps', N') \<Longrightarrow> (\<exists>xs r s.
  w = (xs, r, s) \<and> ps' = (r, s) # ps \<and>
  ws' = ws @ remdups' (norms o snd) (filter (\<lambda>(_, p). norms p \<notin> N)
   (map (\<lambda>a. (a#xs, \<guillemotleft>lderiv a r\<guillemotright>, \<guillemotleft>lderiv a s\<guillemotright>)) (\<sigma> n))) \<and>
  N' = set (map (\<lambda>a. (norm \<guillemotleft>lderiv a r\<guillemotright>, norm \<guillemotleft>lderiv a s\<guillemotright>)) (\<sigma> n)) \<union> N)"
  by (auto split: prod.splits dest!: mp_remdups'
  simp: Let_def norms_def filter_map set_n_lists image_Collect image_image comp_def)

definition closure where "closure n = while_option test (step n)"
definition closure' where "closure' n = while_option test' (step' n)"

definition pre_bisim where
  "pre_bisim n r s = (\<lambda>(ws, ps, N).
     (\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>) \<in> snd ` set ws \<union> set ps \<and>
     distinct (map snd ws @ ps) \<and>
     bij_betw norms (set (map snd ws @ ps)) N \<and>
     (\<forall>(w, r', s') \<in> set ws. \<guillemotleft>lderivs (rev w) r\<guillemotright> = r' \<and> \<guillemotleft>lderivs (rev w) s\<guillemotright> = s' \<and>
        wf_word n (rev w) \<and> wf n r' \<and> wf n s') \<and>
     (\<forall>(r', s') \<in> set ps. (\<exists>w. \<guillemotleft>lderivs w r\<guillemotright> = r' \<and> \<guillemotleft>lderivs w s\<guillemotright> = s') \<and>
       wf n r' \<and> wf n s' \<and> (final r' \<longleftrightarrow> final s') \<and>
       (\<forall>a\<in>set (\<sigma> n). (norm \<guillemotleft>lderiv a r'\<guillemotright>, norm \<guillemotleft>lderiv a s'\<guillemotright>) \<in> N)))"

lemma pre_bisim_start:
  "\<lbrakk>wf n r; wf n s\<rbrakk> \<Longrightarrow> pre_bisim n r s ([([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], [], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)})"
  by (auto simp add: pre_bisim_def bij_betw_def norms_def)

lemma step_mono:
  assumes "step n (ws, ps, N) = (ws', ps', N')"
  shows "snd ` set ws \<union> set ps \<subseteq> snd ` set ws' \<union> set ps'"
  using assms proof (intro subsetI, elim UnE)
  fix x assume "x \<in> snd `set ws"
  with assms show "x \<in> snd ` set ws' \<union> set ps'"
  proof (cases "x = snd (hd ws)")
    case False with `x \<in> image snd (set ws)` have "x \<in> snd ` set (tl ws)" by (cases ws) auto
    with assms show ?thesis by (auto split: prod.splits simp: Let_def)
  qed (auto split: prod.splits simp: Let_def)
qed (auto split: prod.splits simp: Let_def)

lemma pre_bisim_step: "pre_bisim n r s st \<Longrightarrow> test st \<Longrightarrow> pre_bisim n r s (step n st)"
proof (unfold pre_bisim_def, (split prod.splits)+, elim prod_caseE conjE, clarify, intro allI impI conjI)
  fix ws ps N ws' ps' N'
  assume test: "test (ws, ps, N)"
  and step: "step n (ws, ps, N) = (ws', ps', N')"
  and rs: "(\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>) \<in> snd ` set ws \<union> set ps"
  and distinct: "distinct (map snd ws @ ps)"
  and bij: "bij_betw norms (set (map snd ws @ ps)) N"
  and ws: "\<forall>(w, r', s') \<in> set ws. \<guillemotleft>lderivs (rev w) r\<guillemotright> = r' \<and> \<guillemotleft>lderivs (rev w) s\<guillemotright> = s' \<and>
    wf_word n (rev w) \<and> wf n r' \<and> wf n s'"
      (is "\<forall>(w, r', s') \<in> set ws. ?ws w r' s'")
  and ps: "\<forall>(r', s') \<in> set ps. (\<exists>w. \<guillemotleft>lderivs w r\<guillemotright> = r' \<and> \<guillemotleft>lderivs w s\<guillemotright> = s') \<and>
    wf n r' \<and> wf n s' \<and> (final r' \<longleftrightarrow> final s') \<and>
    (\<forall>a\<in>set (\<sigma> n). (norm \<guillemotleft>lderiv a r'\<guillemotright>, norm \<guillemotleft>lderiv a s'\<guillemotright>) \<in> N)"
      (is "\<forall>(r, s) \<in> set ps. ?ps r s N")
  from test obtain x xs where ws_Cons: "ws = x#xs" by (cases ws) auto
  obtain w r' s' where x: "x = (w, r', s')" and ps': "ps' = (r', s') # ps"
    and ws': "ws' = xs @ remdups' (norms o snd) (filter (\<lambda>(_, p). norms p \<notin> N)
      (map (\<lambda>a. (a # w, \<guillemotleft>lderiv a r'\<guillemotright>,  \<guillemotleft>lderiv a s'\<guillemotright>)) (\<sigma> n)))"
    and N': "N' = (set (map (\<lambda>a. (norm \<guillemotleft>lderiv a r'\<guillemotright>, norm \<guillemotleft>lderiv a s'\<guillemotright>)) (\<sigma> n)) - N) \<union> N"
    using step_unfold[OF step[unfolded ws_Cons]] by blast
  hence ws'ps': "set (map snd ws' @ ps') = set (remdups' norms (filter (\<lambda>p. norms p \<notin> N)
    (map (\<lambda>a. (\<guillemotleft>lderiv a r'\<guillemotright>,  \<guillemotleft>lderiv a s'\<guillemotright>)) (\<sigma> n)))) \<union> (set (map snd ws @ ps))"
    unfolding ws' ps' ws_Cons x by (auto dest!: mp_remdups' simp: filter_map image_image image_Un o_def)
  from rs step show "(\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>) \<in> snd ` set ws' \<union> set ps'" by (blast dest: step_mono)
(*next*)
  from distinct ps' ws' ws_Cons x bij show "distinct (map snd ws' @ ps')"
    by (auto simp: bij_betw_def
      intro!: imageI[of _ _ norms] distinct_remdups'_strong
      dest!: mp_remdups'
      elim: image_eqI[of _ snd, OF sym[OF snd_conv]])
(*next*)
  from ps' ws' N' ws x bij show "bij_betw norms (set (map snd ws' @ ps')) N'"
  unfolding ws'ps' N' by (intro bij_betw_combine[OF _ bij]) (auto simp: bij_betw_def norms_def)
(*next*)
  from ws x ws_Cons have wr's': "?ws w r' s'" by auto
  with ws ws_Cons show "\<forall>(w, r', s') \<in> set ws'. ?ws w r' s'" unfolding ws'
    by (auto dest!: mp_remdups' simp: ACI_norm_lderiv elim!: set_mp)
(*next*)
  from ps wr's' test[unfolded ws_Cons x] show "\<forall>(r', s') \<in> set ps'. ?ps r' s' N'" unfolding ps' N'
    by (fastforce simp: image_Collect)
qed

lemma step_commute: "ws \<noteq> [] \<Longrightarrow> (case step n (ws, ps, N) of (ws', ps', N') \<Rightarrow> (map snd ws', N')) = step' n (map snd ws, N)"
apply (auto split: prod.splits)
apply (auto simp only: step.simps step'.simps Let_def map_apfst_remdups' filter_map List.map.compositionality[unfolded comp_def] apfst_def map_pair_def snd_conv id_def)
apply (auto simp: filter_map comp_def map_tl hd_map)
apply (intro image_eqI, auto)+
done

lemma closure_closure':
  "Option.map (\<lambda>(ws, ps, N). (map snd ws, N)) (closure n (ws, ps, N)) =
    closure' n (map snd ws, N)"
  unfolding closure_def closure'_def
  by (rule trans[OF while_option_commute[of _ test' _ _ "step' n"]])
   (auto split: list.splits simp del: step.simps step'.simps List.map.simps simp: step_commute)

theorem closure_sound:
assumes result: "closure n ([([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], [], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}) = Some([], ps, N)"
and wf:  "wf n r" "wf n s"
shows "lang n r = lang n s"
proof -
  from pre_bisim_step pre_bisim_start[OF wf] have pre_bisim_ps: "pre_bisim n r s ([], ps, N)"
    by (rule while_option_rule[OF _ result[unfolded closure_def]])
  then have "is_nbisimulation n (set ps)" "(\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>) \<in> set ps"
    by (auto simp: bij_betw_def pre_bisim_def is_nbisimulation_def in_lists_conv_set norms_def)
  hence "lang n \<guillemotleft>r\<guillemotright> = lang n \<guillemotleft>s\<guillemotright>"
    by (intro nbisim_lang_eq image_eqI) auto
  thus "lang n r = lang n s" unfolding ACI_norm_lang .
qed

theorem closure'_sound:
assumes result: "closure' n ([(\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}) = Some([], N)"
and wf:  "wf n r" "wf n s"
shows "lang n r = lang n s"
  using wf trans[OF closure_closure'[of n "[([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)]" "[]" "{(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}", simplified]
                  result, unfolded option_map_eq_Some]
  by (auto dest: closure_sound)

theorem closure_termination:
assumes wf: "wf n r" "wf n s"
and cl: "closure n ([([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], [], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}) = None" (is "?cl = None")
shows "False"
proof -
  let ?D =  "{norm \<guillemotleft>lderivs xs r\<guillemotright> | xs . True} \<times> {norm \<guillemotleft>lderivs xs s\<guillemotright> | xs . True}"
  let ?X = "\<lambda>ps. ?D - norms ` set ps"
  let ?f = "\<lambda>(ws, ps, N). card (?X ps)"
  have "\<exists>st. ?cl = Some st" unfolding closure_def
  proof (rule measure_while_option_Some[of "pre_bisim n r s" _ _ ?f], intro conjI)
    fix st assume pre_bisim: "pre_bisim n r s st" and "test st"
    hence pre_bisim_step: "pre_bisim n r s (step n st)" by (rule pre_bisim_step)
    obtain ws ps N where st: "st = (ws, ps, N)" by (cases st) blast
    hence "finite (?X ps)" by (blast intro: finite_cartesian_product finite_norm)
    moreover obtain ws' ps' N' where step: "step n (ws, ps, N) = (ws', ps', N')"
      by (cases "step n (ws, ps, N)") blast
    moreover
    { have "norms ` set ps \<subseteq> ?D" using pre_bisim[unfolded st pre_bisim_def]
        by (auto simp: norms_def ACI_norm_lderivs)
      moreover
      have "norms ` set ps' \<subseteq> ?D" using pre_bisim_step[unfolded st step pre_bisim_def]
        by (auto simp: norms_def ACI_norm_lderivs)
      moreover
      { have "distinct (map snd ws @ ps)" "inj_on norms (set (map snd ws @ ps))"
          using pre_bisim[unfolded st pre_bisim_def] by (auto simp: bij_betw_def)
        hence "distinct (map norms (map snd ws @ ps))" unfolding distinct_map ..
        hence "norms ` set ps \<subset> norms ` set (snd (hd ws) # ps)" using `test st` st
          by (cases ws) auto
        moreover have "norms ` set ps' = norms ` set (snd (hd ws) # ps)"
          using step by (auto split: prod.splits)
        ultimately have "norms ` set ps \<subset> norms ` set ps'" by simp
      }
      ultimately have "?X ps' \<subset> ?X ps" by (auto simp add: image_set simp del: set_map)
    }
    ultimately show "?f (step n st) < ?f st" unfolding st step
      using psubset_card_mono[of "?X ps" "?X ps'"] by simp
  qed (auto simp add: pre_bisim_start[OF wf] pre_bisim_step)
  thus False using cl by auto
qed

theorem closure'_termination:
assumes wf: "wf n r" "wf n s"
and cl: "closure' n ([(\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}) = None"
shows "False"
  using wf trans[OF closure_closure'[of n "[([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)]" "[]" "{(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}", simplified]
                  cl, unfolded option_map_is_None]
  by (auto intro: closure_termination)

theorem closure_complete:
assumes eq: "lang n r = lang n s"
and wf:  "wf n r" "wf n s"
shows "\<exists>ps N. closure n ([([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], [], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}) = Some([], ps, N)"
  (is "\<exists>_ _. ?cl = _")
proof (cases ?cl)
  case (Some st)
  moreover obtain ws ps N where ws_ps_N: "st = (ws, ps, N)" by (cases st) blast
  ultimately show ?thesis
  proof (cases ws)
    case (Cons wrs ws)
    then obtain w r' s' where wrs: "wrs = (w, r', s')" by (cases wrs) blast
    with ws_ps_N Cons have "final r' \<noteq> final s'"
      using while_option_stop2[OF Some[unfolded closure_def]] by simp
    moreover
    from pre_bisim_step pre_bisim_start[OF wf] have pre_bisim_ps: "pre_bisim n r s st"
      by (rule while_option_rule[OF _ Some[unfolded closure_def]])
    hence "\<guillemotleft>lderivs (rev w) r\<guillemotright> = r'" "\<guillemotleft>lderivs (rev w) s\<guillemotright> = s'" "wf_word n (rev w)"
      unfolding ws_ps_N Cons wrs pre_bisim_def ACI_norm_lderivs by auto
    ultimately show ?thesis using eq wf lderivs_final by auto
  qed blast
qed (auto intro: closure_termination[OF wf])

theorem closure'_complete:
assumes eq: "lang n r = lang n s"
and wf:  "wf n r" "wf n s"
shows "\<exists>N. closure' n ([(\<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)], {(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}) = Some([], N)"
  using assms closure_closure'[of n "[([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)]" "[]" "{(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}", symmetric]
  by (auto dest!: closure_complete)

text {* The overall procedure *}

definition check_eqv where
"check_eqv n r s \<longleftrightarrow> wf n r \<and> wf n s \<and>
  (let r' = \<guillemotleft>r\<guillemotright>; s' = \<guillemotleft>s\<guillemotright> in (case closure n ([([], r', s')], [], {(norm r', norm s')}) of
     Some([],_) \<Rightarrow> True | _ \<Rightarrow> False))"

definition check_eqv_counterexample where
"check_eqv_counterexample n r s =
  (let r' = \<guillemotleft>r\<guillemotright>; s' = \<guillemotleft>s\<guillemotright> in (case closure n ([([], r', s')], [], {(norm r', norm s')}) of
     Some([],_) \<Rightarrow> None | Some((w,_,_) # _, _) \<Rightarrow> Some w))"

definition check_eqv' where
"check_eqv' n r s \<longleftrightarrow> wf n r \<and> wf n s \<and>
  (let r' = \<guillemotleft>r\<guillemotright>; s' = \<guillemotleft>s\<guillemotright> in (case closure' n ([(r', s')], {(norm r', norm s')}) of
     Some([],_) \<Rightarrow> True | _ \<Rightarrow> False))"

lemma check_eqv_check_eqv': "check_eqv n r s = check_eqv' n r s"
  unfolding check_eqv_def check_eqv'_def Let_def
  using closure_closure'[of n "[([], \<guillemotleft>r\<guillemotright>, \<guillemotleft>s\<guillemotright>)]" "[]" "{(norm \<guillemotleft>r\<guillemotright>, norm \<guillemotleft>s\<guillemotright>)}", symmetric]
  by (auto split: option.splits list.splits)

lemma soundness:
assumes "check_eqv n r s"
shows "lang n r = lang n s"
  using closure_sound assms by (auto simp: check_eqv_def Let_def split: option.splits list.splits)

lemma soundness':
assumes "check_eqv' n r s"
shows "lang n r = lang n s"
  using soundness check_eqv_check_eqv' assms by auto

lemma completeness:
assumes "lang n r = lang n s" "wf n r" "wf n s"
shows "check_eqv n r s"
  using closure_complete[OF assms] assms(2,3) by (auto simp: check_eqv_def)

lemma completeness':
assumes "lang n r = lang n s" "wf n r" "wf n s"
shows "check_eqv' n r s"
  using completeness check_eqv_check_eqv' assms by auto

end

(*
locale efficient_equivalence_checker = 
  wnorm_equivalence_checker \<sigma> \<pi> \<epsilon> norm "project.lang (set o \<sigma>) \<pi>"
  for \<sigma> :: "nat \<Rightarrow> 'a :: linorder list"
  and \<pi> :: "'a \<Rightarrow> 'a"
  and \<epsilon> :: "'a \<Rightarrow> 'a list"
  and norm :: "'a rexp \<Rightarrow> 'a rexp"
begin

fun step'' where "step'' n (ws, N) =
    (let
      (r, s) = hd ws;
      succs = map (\<lambda>a. (norm (lderiv a r), norm (lderiv a s))) (\<sigma> n);
      new = remdups (filter (\<lambda>rs. rs \<notin> N) succs)
    in (new @ tl ws, set new \<union> N))"

definition closure'' where "closure'' n = while_option test' (step'' n)"

definition check_eqv'' where
"check_eqv'' n r s \<longleftrightarrow> wf n r \<and> wf n s \<and>
  (case closure'' n (let
     r' = norm r;
     s' = norm s
     in ([(r', s')], {(r', s')})) of
  Some([],_) \<Rightarrow> True | _ \<Rightarrow> False)"

end
*)
end