(* Author: Dmitriy Traytel *)

header {* Monadic Second-Order Logic Formulas *}

(*<*)
theory Formulas
imports Regular_Operators List_More "~~/src/HOL/Library/RBT_Set"
begin

hide_const (open) RBT.map RBT.fold
(*>*)

subsection {* Interpretations and Encodings *}

type_synonym 'a interp = "'a list \<times> (nat + nat set) list"

abbreviation "enc_atom_bool I n \<equiv> map (\<lambda>x. case x of Inl p \<Rightarrow> n = p | Inr P \<Rightarrow> n \<in> P) I"

abbreviation "enc_atom I n a \<equiv> (a, enc_atom_bool I n)"

subsection {* Syntax and Semantics of MSO *}

datatype 'a formula =
  FQ 'a nat
| FLess nat nat
| FIn nat nat
| FNot "'a formula"
| FOr "'a formula" "'a formula"
| FAnd "'a formula" "'a formula"
| FExists "'a formula"
| FEXISTS "'a formula"

primrec FOV :: "'a formula \<Rightarrow> nat set" where
  "FOV (FQ a m) = {m}"
| "FOV (FLess m1 m2) = {m1, m2}"
| "FOV (FIn m M) = {m}"
| "FOV (FNot \<phi>) = FOV \<phi>"
| "FOV (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) = FOV \<phi>\<^isub>1 \<union> FOV \<phi>\<^isub>2"
| "FOV (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) = FOV \<phi>\<^isub>1 \<union> FOV \<phi>\<^isub>2"
| "FOV (FExists \<phi>) = (\<lambda>x. x - 1) ` (FOV \<phi> - {0})"
| "FOV (FEXISTS \<phi>) = (\<lambda>x. x - 1) ` FOV \<phi>"

primrec SOV :: "'a formula \<Rightarrow> nat set" where
  "SOV (FQ a m) = {}"
| "SOV (FLess m1 m2) = {}"
| "SOV (FIn m M) = {M}"
| "SOV (FNot \<phi>) = SOV \<phi>"
| "SOV (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) = SOV \<phi>\<^isub>1 \<union> SOV \<phi>\<^isub>2"
| "SOV (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) = SOV \<phi>\<^isub>1 \<union> SOV \<phi>\<^isub>2"
| "SOV (FExists \<phi>) = (\<lambda>x. x - 1) ` SOV \<phi>"
| "SOV (FEXISTS \<phi>) = (\<lambda>x. x - 1) ` (SOV \<phi> - {0})"

definition "\<sigma> = (\<lambda>\<Sigma> n. concat (map (\<lambda>bs. map (\<lambda>a. (a, bs)) \<Sigma>) (List.n_lists n [True, False])))"
definition "\<pi> = (\<lambda>(a, bs). (a, tl bs))"
definition "\<epsilon> = (\<lambda>\<Sigma> (a::'a, bs). if a \<in> set \<Sigma> then [(a, True # bs), (a, False # bs)] else [])"

locale formula = embed "set o (\<sigma> \<Sigma>)" \<pi> "\<epsilon> \<Sigma>" for \<Sigma> :: "'a :: linorder list" +
assumes nonempty: "\<Sigma> \<noteq> []"
begin

abbreviation "\<Sigma>_combinatorial_product n \<equiv>
   List.maps (\<lambda>bools. map (\<lambda>a. (a, bools)) \<Sigma>) (bool_combinatorial_product n)"

(* Normal form: quantified variables are used in the body *)
primrec pre_wf_formula :: "nat \<Rightarrow> 'a formula \<Rightarrow> bool" where
  "pre_wf_formula n (FQ a m) = (a \<in> set \<Sigma> \<and> m < n)"
| "pre_wf_formula n (FLess m1 m2) = (m1 < n \<and> m2 < n)"
| "pre_wf_formula n (FIn m M) = (m < n \<and> M < n)"
| "pre_wf_formula n (FNot \<phi>) = pre_wf_formula n \<phi>"
| "pre_wf_formula n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) = (pre_wf_formula n \<phi>\<^isub>1 \<and> pre_wf_formula n \<phi>\<^isub>2)"
| "pre_wf_formula n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) = (pre_wf_formula n \<phi>\<^isub>1 \<and> pre_wf_formula n \<phi>\<^isub>2)"
| "pre_wf_formula n (FExists \<phi>) = (pre_wf_formula (n + 1) \<phi> \<and> 0 \<in> FOV \<phi> \<and> 0 \<notin> SOV \<phi>)"
| "pre_wf_formula n (FEXISTS \<phi>) = (pre_wf_formula (n + 1) \<phi> \<and> 0 \<notin> FOV \<phi> \<and> 0 \<in> SOV \<phi>)"

abbreviation "closed \<equiv> pre_wf_formula 0"

definition [simp]: "wf_formula n \<phi> \<equiv> pre_wf_formula n \<phi> \<and> FOV \<phi> \<inter> SOV \<phi> = {}"

lemma max_idx_vars: "pre_wf_formula n \<phi> \<Longrightarrow> \<forall>p \<in> FOV \<phi> \<union> SOV \<phi>. p < n"
  by (induct \<phi> arbitrary: n)
   (auto split: split_if_asm, (metis Un_iff diff_Suc_less less_SucE less_imp_diff_less)+)

lemma finite_FOV: "finite (FOV \<phi>)"
  by (induct \<phi>) (auto split: split_if_asm)

subsection {* ENC *}

definition "arbitrary_except n pbs xs \<equiv>
   PLUS (map (\<lambda>bs. PLUS (map (\<lambda>x. Atom (x, fold (\<lambda>(p, b). insert_nth p b) pbs bs)) xs))
    (bool_combinatorial_product (n - length pbs)))"

lemma wf_rexp_arbitrary_except:
  "\<lbrakk>length pbs \<le> n; set xs \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow> wf n (arbitrary_except n pbs xs)"
  by (auto simp: arbitrary_except_def) (force simp: \<sigma>_def set_n_lists length_fold_insert_nth)

definition valid_ENC :: "nat \<Rightarrow> nat \<Rightarrow> ('a \<times> bool list) rexp" where
  "valid_ENC n p = (if n = 0 then Regular_Exp.Not Zero else
    TIMES [
      Star (arbitrary_except n [(p, False)] \<Sigma>),
      arbitrary_except n [(p, True)] \<Sigma>,
      Star (arbitrary_except n [(p, False)] \<Sigma>)])"

lemma wf_rexp_valid_ENC: "wf n (valid_ENC n p)"
  unfolding valid_ENC_def by (auto intro!: wf_rexp_arbitrary_except)

definition ENC :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "ENC n \<phi> = flatten INTERSECT (valid_ENC n ` FOV \<phi>)"

lemma wf_rexp_ENC: "wf n (ENC n \<phi>)"
  unfolding ENC_def
  by (intro iffD2[OF wf_flatten_INTERSECT]) (auto intro: finite_FOV simp: wf_rexp_valid_ENC)

lemma enc_atom_\<sigma>_eq: "i < length w \<Longrightarrow>
 (length I = n \<and> w ! i \<in> set \<Sigma>) \<longleftrightarrow> enc_atom I i (w ! i) \<in> set (\<sigma> \<Sigma> n)"
  by (auto simp: \<sigma>_def set_n_lists intro!: exI[of _ "enc_atom_bool I i"] imageI)

lemmas enc_atom_\<sigma> = iffD1[OF enc_atom_\<sigma>_eq, OF _ conjI]

lemma enc_atom_bool_take_drop_True:
  "\<lbrakk>r < length I; case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P\<rbrakk> \<Longrightarrow>
    enc_atom_bool I p = take r (enc_atom_bool I p) @ True # drop (Suc r) (enc_atom_bool I p)"
  by (auto intro: trans[OF id_take_nth_drop])

lemma enc_atom_bool_take_drop_True2:
  "\<lbrakk>r < length I; case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P;
    s < length I; case I ! s of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P; r < s\<rbrakk> \<Longrightarrow>
    enc_atom_bool I p = take r (enc_atom_bool I p) @ True #
      take (s - Suc r) (drop (Suc r) (enc_atom_bool I p)) @ True #
      drop (Suc s) (enc_atom_bool I p)"
  using id_take_nth_drop[of r "enc_atom_bool I p"]
      id_take_nth_drop[of "s - r - 1" "drop (Suc r) (enc_atom_bool I p)"]
  by auto

lemma enc_atom_bool_take_drop_False:
  "\<lbrakk>r < length I; case I ! r of Inl p' \<Rightarrow> p \<noteq> p' | Inr P \<Rightarrow> p \<notin> P\<rbrakk> \<Longrightarrow>
    enc_atom_bool I p = take r (enc_atom_bool I p) @ False # drop (Suc r) (enc_atom_bool I p)"
  by (auto intro: trans[OF id_take_nth_drop] split: sum.splits)

lemma enc_atom_lang_arbitrary_except_True: "\<lbrakk>r < length I;
   case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P; length I = n; a \<in> set xs; set xs \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
  [enc_atom I p a] \<in> lang n (arbitrary_except n [(r, True)] xs)"
  unfolding arbitrary_except_def
  by (auto intro!: bexI[of _ "(\<lambda>J. take r J @ drop (r + 1) J) (enc_atom_bool I p)"]
         intro: enc_atom_bool_take_drop_True)

lemma enc_atom_lang_arbitrary_except_True2:"\<lbrakk>r < length I; s < length I; r < s; 
  case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P;
  case I ! s of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P; length I = n; a \<in> set \<Sigma>\<rbrakk> \<Longrightarrow>
  [enc_atom I p a] \<in> lang n (arbitrary_except n [(r, True), (s, True)] \<Sigma>)"
  unfolding arbitrary_except_def
  by (auto intro!: bexI[of _ "(\<lambda>J. take r J @ take (s - r - 1) (drop (r + 1) J) @ drop (s - r) (drop (r + 1) J)) (enc_atom_bool I p)"]
         intro!: enc_atom_bool_take_drop_True2 simp: min_absorb2 take_Cons' drop_Cons')

lemma enc_atom_lang_arbitrary_except_False:"\<lbrakk>r < length I;
  case I ! r of Inl p' \<Rightarrow> p \<noteq> p' | Inr P \<Rightarrow> p \<notin> P; length I = n; a \<in> set xs; set xs \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
  [enc_atom I p a] \<in> lang n (arbitrary_except n [(r, False)] xs)"
  unfolding arbitrary_except_def
  by (auto intro!: bexI[of _ "(\<lambda>J. take r J @ drop (r + 1) J) (enc_atom_bool I p)"]
         intro: enc_atom_bool_take_drop_False)

lemma arbitrary_except:
  "\<lbrakk>v \<in> lang n (arbitrary_except n [(r, b)] S); r < n; set S \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
    \<exists>x. v = [x] \<and> snd x ! r = b \<and> fst x \<in> set S"
  unfolding arbitrary_except_def by (auto simp: nth_append)

lemma arbitrary_except2:
  "\<lbrakk>v \<in> lang n (arbitrary_except n [(r, b), (s, b')] S); r < s; r < n; s < n; set S \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
    \<exists>x. v = [x] \<and> snd x ! r = b \<and> snd x ! s = b' \<and> fst x \<in> set S"
  unfolding arbitrary_except_def by (auto simp: nth_append min_absorb2)

lemma star_arbitrary_except:
  "\<lbrakk>v \<in> star (lang n (arbitrary_except n [(r, b)] \<Sigma>)); r < n; i < length v\<rbrakk> \<Longrightarrow>
    snd (v ! i) ! r = b"
proof (induct arbitrary: i rule: star_induct)
  case (append u v) thus ?case by (cases i) (auto dest: arbitrary_except)
qed simp

end

end
