(* Author: Dmitriy Traytel *)

header {* Monadic Second-Order Logic on Finite Words *}

(*<*)
theory MSO
imports Regular_Exp "~~/src/HOL/Library/Product_ord"  "~~/src/HOL/Library/RBT_Set"
begin

declare set.simps [code]
hide_const (open) RBT.map RBT.fold
(*>*)

subsection {* Library Functions *}

primrec map_index' :: "nat \<Rightarrow> (nat \<Rightarrow> 'a \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b list" where
  "map_index' n f [] = []"
| "map_index' n f (x#xs) = f n x # map_index' (Suc n) f xs"

lemma length_map_index'[simp]: "length (map_index' n f xs) = length xs"
by (induct xs arbitrary: n) auto

lemma map_index'_map_zip: "map_index' n f xs = map (split f) (zip [n ..< n + length xs] xs)"
proof (induct xs arbitrary: n)
  case (Cons x xs)
  hence "map_index' n f (x#xs) = f n x # map (split f) (zip [Suc n ..< n + length (x # xs)] xs)" by simp
  also have "\<dots> =  map (split f) (zip (n # [Suc n ..< n + length (x # xs)]) (x # xs))" by simp
  also have "(n # [Suc n ..< n + length (x # xs)]) = [n ..< n + length (x # xs)]" by (induct xs) auto
  finally show ?case by simp
qed simp

abbreviation "map_index \<equiv> map_index' 0"

lemmas map_index = map_index'_map_zip[of 0, simplified]

lemma take_map_index: "take p (map_index f xs) = map_index f (take p xs)"
unfolding map_index by (auto simp: min_def take_map take_zip)

lemma drop_map_index: "drop p (map_index f xs) = map_index' p f (drop p xs)"
unfolding map_index'_map_zip by (cases "p < length xs") (auto simp: drop_map drop_zip)

lemma map_map_index[simp]: "map g (map_index f xs) = map_index (\<lambda>n x. g (f n x)) xs"
unfolding map_index by auto

lemma map_index_map[simp]: "map_index f (map g xs) = map_index (\<lambda>n x. f n (g x)) xs"
unfolding map_index by (auto simp: map_zip_map2)

lemma set_map_index[simp]: "x \<in> set (map_index f xs) = (\<exists>i < length xs. f i (xs ! i) = x)"
unfolding map_index by (auto simp: set_zip intro!: image_eqI[of _ "split f"])

lemma nth_map_index[simp]: "p < length xs \<Longrightarrow> map_index f xs ! p = f p (xs ! p)"
unfolding map_index by auto

lemma map_index_cong:
  "\<forall>p < length xs. f p (xs ! p) = g p (xs ! p) \<Longrightarrow> map_index f xs = map_index g xs"
unfolding map_index by (auto simp: set_zip)

lemma map_index_id: "map_index (curry snd) xs = xs"
unfolding map_index by auto

lemma map_index_no_index[simp]: "map_index (\<lambda>n x. f x) xs = map f xs"
unfolding map_index by (induct xs rule: rev_induct) auto

lemma map_index_congL:
  "\<forall>p < length xs. f p (xs ! p) = xs ! p \<Longrightarrow> map_index f xs = xs"
by (rule trans[OF map_index_cong map_index_id]) auto

lemma map_index'_is_NilD: "map_index' n f xs = [] \<Longrightarrow> xs = []"
by (induct xs) auto

declare map_index'_is_NilD[of 0, dest!]

lemma map_index'_is_ConsD:
  "map_index' n f xs = y # ys \<Longrightarrow> \<exists>z zs. xs = z # zs \<and> f n z = y \<and> map_index' (n + 1) f zs = ys"
by (induct xs arbitrary: n) auto

lemma map_index'_eq_imp_length_eq: "map_index' n f xs = map_index' n g ys \<Longrightarrow> length xs = length ys"
proof (induct ys arbitrary: xs n)
  case (Cons y ys) thus ?case by (cases xs) auto
qed (auto dest!: map_index'_is_NilD)

lemmas map_index_eq_imp_length_eq = map_index'_eq_imp_length_eq[of 0]

lemma map_index'_comp[simp]: "map_index' n f (map_index' n g xs) = map_index' n (\<lambda>n. f n o g n) xs"
by (induct xs arbitrary: n) auto

primrec combinatorial_product :: "'a list list \<Rightarrow> 'a list list" where
  "combinatorial_product [] = [[]]"
| "combinatorial_product (xs # xss) = List.maps (\<lambda>x. map (Cons x) (combinatorial_product xss)) xs"

abbreviation "bool_combinatorial_product n \<equiv> combinatorial_product (replicate n [True, False])"

lemma set_bool_combinatorial_product[simp]:
  "bs \<in> set (bool_combinatorial_product n) \<longleftrightarrow> length bs = n"
proof (induct bs arbitrary: n)
  case Nil thus ?case by (cases n) (auto simp: maps_def)
next
  case (Cons b bs) thus ?case by (cases n) (auto simp: maps_def)
qed

fun insert_nth :: "nat \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "insert_nth 0 x xs = x # xs"
| "insert_nth (Suc n) x (y # xs) = y # insert_nth n x xs"
| "insert_nth _ x [] = [x]"

lemma insert_nth_take_drop[simp]: "insert_nth n x xs = take n xs @ [x] @ drop n xs"
by (induct n x xs rule: insert_nth.induct) auto

lemma length_insert_nth: "length (insert_nth n x xs) = Suc (length xs)"
by (induct xs) auto

subsection {* Interpretations and Encodings *}

type_synonym 'a interp = "'a list \<times> (nat + nat set) list"

abbreviation "enc_atom_bool I n \<equiv> map (\<lambda>x. case x of Inl p \<Rightarrow> n = p | Inr P \<Rightarrow> n \<in> P) I"

abbreviation "enc_atom I n a \<equiv> (a, enc_atom_bool I n)"

fun enc :: "'a interp \<Rightarrow> ('a \<times> bool list) list" where
  "enc (w, I) = map_index (enc_atom I) w"

subsection {* Syntax and Semantics of MSO *}

datatype 'a formula =
  Q 'a nat
| Less nat nat
| In nat nat
| Not "'a formula"
| Or "'a formula" "'a formula"
| And "'a formula" "'a formula"
| Exists "'a formula"
| EXISTS "'a formula"

definition "\<sigma> = (\<lambda>\<Sigma> n. concat (map (\<lambda>bs. map (\<lambda>a. (a, bs)) \<Sigma>) (List.n_lists n [True, False])))"
definition "\<pi> = (\<lambda>(a, bs). (a, tl bs))"

locale MSO = project "set o (\<sigma> \<Sigma>)" \<pi> for \<Sigma> :: "'a :: linorder list"
begin

abbreviation "\<Sigma>_combinatorial_product n \<equiv>
   List.maps (\<lambda>bools. map (\<lambda>a. (a, bools)) \<Sigma>) (bool_combinatorial_product n)"

primrec FOV :: "'a formula \<Rightarrow> nat set" where
  "FOV (Q a m) = {m}"
| "FOV (Less m1 m2) = {m1, m2}"
| "FOV (In m M) = {m}"
| "FOV (Not \<phi>) = FOV \<phi>"
| "FOV (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = FOV \<phi>\<^isub>1 \<union> FOV \<phi>\<^isub>2"
| "FOV (And \<phi>\<^isub>1 \<phi>\<^isub>2) = FOV \<phi>\<^isub>1 \<union> FOV \<phi>\<^isub>2"
| "FOV (Exists \<phi>) = (\<lambda>x. x - 1) ` (FOV \<phi> - {0})"
| "FOV (EXISTS \<phi>) = (\<lambda>x. x - 1) ` FOV \<phi>"

primrec SOV :: "'a formula \<Rightarrow> nat set" where
  "SOV (Q a m) = {}"
| "SOV (Less m1 m2) = {}"
| "SOV (In m M) = {M}"
| "SOV (Not \<phi>) = SOV \<phi>"
| "SOV (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = SOV \<phi>\<^isub>1 \<union> SOV \<phi>\<^isub>2"
| "SOV (And \<phi>\<^isub>1 \<phi>\<^isub>2) = SOV \<phi>\<^isub>1 \<union> SOV \<phi>\<^isub>2"
| "SOV (Exists \<phi>) = (\<lambda>x. x - 1) ` SOV \<phi>"
| "SOV (EXISTS \<phi>) = (\<lambda>x. x - 1) ` (SOV \<phi> - {0})"

(* Normal form: quantified variables are used in the body *)
primrec pre_wf_formula :: "nat \<Rightarrow> 'a formula \<Rightarrow> bool" where
  "pre_wf_formula n (Q a m) = (a \<in> set \<Sigma> \<and> m < n)"
| "pre_wf_formula n (Less m1 m2) = (m1 < n \<and> m2 < n)"
| "pre_wf_formula n (In m M) = (m < n \<and> M < n)"
| "pre_wf_formula n (Not \<phi>) = pre_wf_formula n \<phi>"
| "pre_wf_formula n (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = (pre_wf_formula n \<phi>\<^isub>1 \<and> pre_wf_formula n \<phi>\<^isub>2)"
| "pre_wf_formula n (And \<phi>\<^isub>1 \<phi>\<^isub>2) = (pre_wf_formula n \<phi>\<^isub>1 \<and> pre_wf_formula n \<phi>\<^isub>2)"
| "pre_wf_formula n (Exists \<phi>) = (pre_wf_formula (n + 1) \<phi> \<and> 0 \<in> FOV \<phi> \<and> 0 \<notin> SOV \<phi>)"
| "pre_wf_formula n (EXISTS \<phi>) = (pre_wf_formula (n + 1) \<phi> \<and> 0 \<notin> FOV \<phi> \<and> 0 \<in> SOV \<phi>)"

abbreviation "closed \<equiv> pre_wf_formula 0"

abbreviation "wf_formula n \<phi> \<equiv> pre_wf_formula n \<phi> \<and> FOV \<phi> \<inter> SOV \<phi> = {}"

lemma max_idx_vars: "pre_wf_formula n \<phi> \<Longrightarrow> \<forall>p \<in> FOV \<phi> \<union> SOV \<phi>. p < n"
by (induct \<phi> arbitrary: n)
   (auto split: split_if_asm, (metis Un_iff diff_Suc_less less_SucE less_imp_diff_less)+)

lemma finite_FOV: "finite (FOV \<phi>)"
by (induct \<phi>) (auto split: split_if_asm)

lemma finite_SOV: "finite (SOV \<phi>)"
by (induct \<phi>) (auto split: split_if_asm)

abbreviation "wf_interp w I \<equiv> (length w > 0 \<and>
    (\<forall>a \<in> set w. a \<in> set \<Sigma>) \<and>
    (\<forall>x \<in> set I. case x of Inl p \<Rightarrow> p < length w | Inr P \<Rightarrow> \<forall>p \<in> P. p < length w))"

fun wf_interp_for_formula :: "'a interp \<Rightarrow> 'a formula \<Rightarrow> bool" where
  "wf_interp_for_formula (w, I) \<phi> =
    (wf_interp w I \<and>
    (\<forall>n \<in> FOV \<phi>. case I ! n of Inl _ \<Rightarrow> True | _ \<Rightarrow> False) \<and>
    (\<forall>n \<in> SOV \<phi>. case I ! n of Inl _ \<Rightarrow> False | _ \<Rightarrow> True))"

fun satisfies :: "'a interp \<Rightarrow> 'a formula \<Rightarrow> bool" (infix "\<Turnstile>" 50) where
  "(w, I) \<Turnstile> Q a m = (w ! (case I ! m of Inl p \<Rightarrow> p) = a)"
| "(w, I) \<Turnstile> Less m1 m2 = ((case I ! m1 of Inl p \<Rightarrow> p) < (case I ! m2 of Inl p \<Rightarrow> p))"
| "(w, I) \<Turnstile> In m M = ((case I ! m of Inl p \<Rightarrow> p) \<in> (case I ! M of Inr P \<Rightarrow> P))"
| "(w, I) \<Turnstile> Not \<phi> = (\<not> (w, I) \<Turnstile> \<phi>)"
| "(w, I) \<Turnstile> (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = ((w, I) \<Turnstile> \<phi>\<^isub>1 \<or> (w, I) \<Turnstile> \<phi>\<^isub>2)"
| "(w, I) \<Turnstile> (And \<phi>\<^isub>1 \<phi>\<^isub>2) = ((w, I) \<Turnstile> \<phi>\<^isub>1 \<and> (w, I) \<Turnstile> \<phi>\<^isub>2)"
| "(w, I) \<Turnstile> (Exists \<phi>) = (\<exists>p. p \<in> {0 .. length w - 1} \<and> (w, Inl p # I) \<Turnstile> \<phi>)"
| "(w, I) \<Turnstile> (EXISTS \<phi>) = (\<exists>P. P \<subseteq> {0 .. length w - 1} \<and> (w, Inr P # I) \<Turnstile> \<phi>)"

definition lang\<^isub>M\<^isub>S\<^isub>O :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) list set" where
  "lang\<^isub>M\<^isub>S\<^isub>O n \<phi> = {enc (w, I) | w I.
     length I = n \<and> wf_interp_for_formula (w, I) \<phi> \<and> satisfies (w, I) \<phi>}"

subsection {* ENC *}

definition "arbitrary_except n pbs xs \<equiv>
   PLUS (map (\<lambda>bs. PLUS (map (\<lambda>x. Atom (x, fold (\<lambda>(p, b). insert_nth p b) pbs bs)) xs))
    (bool_combinatorial_product (n - length pbs)))"

lemma length_fold_insert_nth:
  "length (fold (\<lambda>(p, b). insert_nth p b) pbs bs) = length bs + length pbs"
by (induct pbs arbitrary: bs) auto

lemma invar_fold_insert_nth:
  "\<lbrakk>\<forall>x\<in>set pbs. p < fst x; p < length bs; bs ! p = b\<rbrakk> \<Longrightarrow>
    fold (\<lambda>(x, y). insert_nth x y) pbs bs ! p = b"
by (induct pbs arbitrary: bs) (auto simp: nth_append)

lemma nth_fold_insert_nth:
  "\<lbrakk>sorted (map fst pbs); distinct (map fst pbs); \<forall>(p, b) \<in> set pbs. p < length bs + length pbs;
    i < length pbs; pbs ! i = (p, b)\<rbrakk> \<Longrightarrow>
  fold (\<lambda>(p, b). insert_nth p b) pbs bs ! p = b"
proof (induct pbs arbitrary: bs i p b)
  case (Cons pb pbs)
  show ?case
  proof (cases i)
    case 0
    with Cons.prems have "p < Suc (length bs)"
    proof (induct pbs rule: rev_induct)
      case (snoc pb' pbs)
      then obtain p' b' where "pb' = (p', b')" by auto
      with snoc.prems have "\<forall>p \<in> fst ` set pbs. p < p'" "p' \<le> Suc (length bs + length pbs)"
        by (auto simp: image_iff sorted_Cons sorted_append le_eq_less_or_eq)
      with snoc.prems show ?case by (intro snoc(1)) (auto simp: sorted_Cons sorted_append)
    qed auto
    with 0 Cons.prems show ?thesis unfolding fold.simps o_apply
    by (intro invar_fold_insert_nth) (auto simp: sorted_Cons image_iff le_eq_less_or_eq nth_append)
  next
    case (Suc n) with Cons.prems show ?thesis unfolding fold.simps
      by (auto intro!: Cons(1) simp: sorted_Cons)
  qed
qed simp

lemma wf_rexp_arbitrary_except:
  "\<lbrakk>length pbs \<le> n; set xs \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow> wf n (arbitrary_except n pbs xs)"
by (auto simp: arbitrary_except_def wf_PLUS) (force simp: \<sigma>_def set_n_lists length_fold_insert_nth)

definition valid_ENC :: "nat \<Rightarrow> nat \<Rightarrow> ('a \<times> bool list) rexp" where
  "valid_ENC n p = (if n = 0 then Regular_Exp.Not Zero else
    TIMES [
      Star (arbitrary_except n [(p, False)] \<Sigma>),
      arbitrary_except n [(p, True)] \<Sigma>,
      Star (arbitrary_except n [(p, False)] \<Sigma>)])"

lemma wf_rexp_valid_ENC: "wf n (valid_ENC n p)"
unfolding valid_ENC_def by (auto intro!: wf_rexp_arbitrary_except)

definition ENC :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "ENC n \<phi> = flatten INTERSECT (valid_ENC n ` FOV \<phi>)"

lemma wf_rexp_ENC: "wf n (ENC n \<phi>)"
unfolding ENC_def
by (intro iffD2[OF wf_flatten_INTERSECT]) (auto intro: finite_FOV simp: wf_rexp_valid_ENC)

lemma enc_atom_\<sigma>_eq: "i < length w \<Longrightarrow>
 (length I = n \<and> w ! i \<in> set \<Sigma>) \<longleftrightarrow> enc_atom I i (w ! i) \<in> set (\<sigma> \<Sigma> n)"
by (auto simp: \<sigma>_def set_n_lists intro!: exI[of _ "enc_atom_bool I i"] imageI)

lemmas enc_atom_\<sigma> = iffD1[OF enc_atom_\<sigma>_eq, OF _ conjI]

lemma enc_atom_bool_take_drop_True:
  "\<lbrakk>r < length I; case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P\<rbrakk> \<Longrightarrow>
    enc_atom_bool I p = take r (enc_atom_bool I p) @ True # drop (Suc r) (enc_atom_bool I p)"
by (auto intro: trans[OF id_take_nth_drop])

lemma enc_atom_bool_take_drop_True2:
  "\<lbrakk>r < length I; case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P;
    s < length I; case I ! s of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P; r < s\<rbrakk> \<Longrightarrow>
    enc_atom_bool I p = take r (enc_atom_bool I p) @ True #
      take (s - Suc r) (drop (Suc r) (enc_atom_bool I p)) @ True #
      drop (Suc s) (enc_atom_bool I p)"
using id_take_nth_drop[of r "enc_atom_bool I p"]
      id_take_nth_drop[of "s - r - 1" "drop (Suc r) (enc_atom_bool I p)"]
by auto

lemma enc_atom_bool_take_drop_False:
  "\<lbrakk>r < length I; case I ! r of Inl p' \<Rightarrow> p \<noteq> p' | Inr P \<Rightarrow> p \<notin> P\<rbrakk> \<Longrightarrow>
    enc_atom_bool I p = take r (enc_atom_bool I p) @ False # drop (Suc r) (enc_atom_bool I p)"
by (auto intro: trans[OF id_take_nth_drop] split: sum.splits)

lemma enc_atom_lang_arbitrary_except_True: "\<lbrakk>r < length I;
   case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P; length I = n; a \<in> set xs; set xs \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
  [enc_atom I p a] \<in> lang n (arbitrary_except n [(r, True)] xs)"
unfolding arbitrary_except_def
by (auto intro!: bexI[of _ "(\<lambda>J. take r J @ drop (r + 1) J) (enc_atom_bool I p)"]
         intro: enc_atom_bool_take_drop_True)

lemma enc_atom_lang_arbitrary_except_True2:"\<lbrakk>r < length I; s < length I; r < s; 
  case I ! r of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P;
  case I ! s of Inl p' \<Rightarrow> p = p' | Inr P \<Rightarrow> p \<in> P; length I = n; a \<in> set \<Sigma>\<rbrakk> \<Longrightarrow>
  [enc_atom I p a] \<in> lang n (arbitrary_except n [(r, True), (s, True)] \<Sigma>)"
unfolding arbitrary_except_def
by (auto intro!: bexI[of _ "(\<lambda>J. take r J @ take (s - r - 1) (drop (r + 1) J) @ drop (s - r) (drop (r + 1) J)) (enc_atom_bool I p)"]
         intro!: enc_atom_bool_take_drop_True2 simp: min_absorb2 take_Cons' drop_Cons')

lemma enc_atom_lang_arbitrary_except_False:"\<lbrakk>r < length I;
  case I ! r of Inl p' \<Rightarrow> p \<noteq> p' | Inr P \<Rightarrow> p \<notin> P; length I = n; a \<in> set xs; set xs \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
  [enc_atom I p a] \<in> lang n (arbitrary_except n [(r, False)] xs)"
unfolding arbitrary_except_def
by (auto intro!: bexI[of _ "(\<lambda>J. take r J @ drop (r + 1) J) (enc_atom_bool I p)"]
         intro: enc_atom_bool_take_drop_False)

lemma arbitrary_except:
  "\<lbrakk>v \<in> lang n (arbitrary_except n [(r, b)] S); r < n; set S \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
    \<exists>x. v = [x] \<and> snd x ! r = b \<and> fst x \<in> set S"
unfolding arbitrary_except_def by (auto simp: nth_append)

lemma arbitrary_except2:
  "\<lbrakk>v \<in> lang n (arbitrary_except n [(r, b), (s, b')] S); r < s; r < n; s < n; set S \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow>
    \<exists>x. v = [x] \<and> snd x ! r = b \<and> snd x ! s = b' \<and> fst x \<in> set S"
unfolding arbitrary_except_def by (auto simp: nth_append min_absorb2)

lemma star_arbitrary_except:
  "\<lbrakk>v \<in> star (lang n (arbitrary_except n [(r, b)] \<Sigma>)); r < n; i < length v\<rbrakk> \<Longrightarrow>
    snd (v ! i) ! r = b"
proof (induct arbitrary: i rule: star_induct)
  case (append u v) thus ?case by (cases i) (auto dest: arbitrary_except)
qed simp

definition "dec_word \<equiv> map fst"

definition "positions_in_row w i =
  Option.these (set (map_index (\<lambda>p a_bs. if nth (snd a_bs) i then Some p else None) w))"

definition "dec_interp n FO w \<equiv> map (\<lambda>i.
  if i \<in> FO
  then Inl (the_elem (positions_in_row w i))
  else Inr (positions_in_row w i)) [0..<n]"

lemma positions_in_row: "positions_in_row w i = {p. p < length w \<and> snd (w ! p) ! i}"
unfolding positions_in_row_def these_def by (auto intro!: image_eqI[of _ the])

lemma positions_in_row_unique: "\<exists>!p. p < length w \<and> snd (w ! p) ! i \<Longrightarrow>
  the_elem (positions_in_row w i) = (THE p. p < length w \<and> snd (w ! p) ! i)"
by (rule the1I2) (auto simp: the_elem_def positions_in_row)

lemma positions_in_row_length: "\<exists>!p. p < length w \<and> snd (w ! p) ! i \<Longrightarrow>
  the_elem (positions_in_row w i) < length w"
unfolding positions_in_row_unique by (rule the1I2) auto

lemma positions_in_row_nth: "\<exists>!p. p < length w \<and> snd (w ! p) ! i \<Longrightarrow>
  snd (w ! the_elem (positions_in_row w i)) ! i"
unfolding positions_in_row_unique by (rule the1I2) auto

lemma dec_interp_Inl: "\<lbrakk>i \<in> FO; i < n\<rbrakk> \<Longrightarrow> \<exists>p. dec_interp n FO x ! i = Inl p"
unfolding dec_interp_def using nth_map[of n "[0..<n]"] by auto

lemma dec_interp_not_Inr: "\<lbrakk>dec_interp n FO x ! i = Inr P; i \<in> FO; i < n\<rbrakk> \<Longrightarrow> False"
unfolding dec_interp_def using nth_map[of n "[0..<n]"] by auto

lemma dec_interp_Inr: "\<lbrakk>i \<notin> FO; i < n\<rbrakk> \<Longrightarrow> \<exists>P. dec_interp n FO x ! i = Inr P"
unfolding dec_interp_def using nth_map[of n "[0..<n]"] by auto

lemma dec_interp_not_Inl: "\<lbrakk>dec_interp n FO x ! i = Inl p; i \<notin> FO; i < n\<rbrakk> \<Longrightarrow> False"
unfolding dec_interp_def using nth_map[of n "[0..<n]"] by auto

lemma Inl_dec_interp_length:
assumes "\<forall>i \<in> FO. \<exists>!p. p < length w \<and> snd (w ! p) ! i"
shows "Inl p \<in> set (dec_interp n FO w) \<Longrightarrow> p < length w"
unfolding dec_interp_def by (auto intro: positions_in_row_length[OF bspec[OF assms]])

lemma Inr_dec_interp_length: "\<lbrakk>Inr P \<in> set (dec_interp n FO w); p \<in> P\<rbrakk> \<Longrightarrow> p < length w"
unfolding dec_interp_def by (auto simp: positions_in_row)

lemma enc_atom_dec: 
  "\<lbrakk>wf_word n w; \<forall>i \<in> FO. i < n \<longrightarrow> (\<exists>!p. p < length w \<and> snd (w ! p) ! i); p < length w\<rbrakk> \<Longrightarrow>
   enc_atom (dec_interp n FO w) p (fst (w ! p)) = w ! p"
unfolding wf_word dec_interp_def map_filter_def Let_def
apply (auto simp: comp_def intro!: trans[OF iffD2[OF Pair_eq] pair_collapse])
apply (intro nth_equalityI)
apply (auto simp add: \<sigma>_def set_n_lists dest!: nth_mem) []
apply (auto simp: positions_in_row)
apply (drule bspec)
apply assumption
apply (drule mp)
apply assumption
apply (frule positions_in_row_unique)
apply (erule notE)
apply (simp add: positions_in_row)
apply (erule the1I2)
apply simp

apply (drule bspec)
apply assumption
apply (drule mp)
apply assumption
apply (frule positions_in_row_unique)
apply (simp add: positions_in_row)
apply (rule the_equality[symmetric])
apply auto
done

lemma enc_dec:
  "\<lbrakk>wf_word n w; \<forall>i \<in> FO. i < n \<longrightarrow> (\<exists>!p. p < length w \<and> snd (w ! p) ! i)\<rbrakk> \<Longrightarrow>
   enc (dec_word w, dec_interp n FO w) = w"
unfolding enc.simps dec_word_def
by (intro trans[OF map_index_map map_index_congL] allI impI enc_atom_dec) assumption+

lemma dec_word_enc: "dec_word (enc (w, I)) = w"
unfolding dec_word_def by auto

lemma enc_unique: 
  assumes "wf_interp w I" "i < length I"
  shows "\<exists>p. I ! i = Inl p \<Longrightarrow> \<exists>!p. p < length (enc (w, I)) \<and> snd (enc (w, I) ! p) ! i"
proof (erule exE)
  fix p assume "I ! i = Inl p"
  with assms show ?thesis by (auto simp: map_index all_set_conv_all_nth intro!: exI[of _ p])
qed

lemma dec_interp_enc_Inl:
  "\<lbrakk>dec_interp n FO (enc (w, I)) ! i = Inl p'; I ! i = Inl p; i \<in> FO; i < n; length I = n; p < length w; wf_interp w I\<rbrakk> \<Longrightarrow>
  p = p'"
unfolding dec_interp_def using nth_map[of _ "[0..<n]"] positions_in_row_unique[OF enc_unique]
by (auto intro: sym[OF the_equality])

lemma dec_interp_enc_Inr:
  "\<lbrakk>dec_interp n FO (enc (w, I)) ! i = Inr P'; I ! i = Inr P; i \<notin> FO; i < n; length I = n; \<forall>p \<in> P. p < length w\<rbrakk> \<Longrightarrow>
  P = P'"
unfolding dec_interp_def positions_in_row by auto

lemma lang_ENC:
 assumes "wf_formula n \<phi>"
 shows "lang n (ENC n \<phi>) - {[]} = {enc (w, I) | w I . length I = n \<and> wf_interp_for_formula (w, I) \<phi>}"
   (is "?L = ?R")
proof (cases "FOV \<phi> = {}")
  case True with assms show ?thesis
    apply (auto simp: ENC_def intro!: enc_atom_\<sigma>)
    apply (intro exI conjI)
    apply (rule trans[OF sym[OF enc_dec[of _ _ "{}"]] enc.simps])
    apply (auto simp add: wf_word dec_word_def dec_interp_def)
    apply (auto simp: \<sigma>_def positions_in_row set_n_lists split: sum.splits)
    apply (drule trans[OF sym nth_map])
    apply (auto intro: bspec[OF max_idx_vars])
  done
next
  case False
  hence nonempty: "valid_ENC n ` FOV \<phi> \<noteq> {}" by simp
  have finite: "finite (valid_ENC n ` FOV \<phi>)" by (rule finite_imageI[OF finite_FOV])
  from False assms(1) have "0 < n" by (cases n) (auto split: dest!: max_idx_vars)
  with wf_rexp_valid_ENC have wf_rexp: "\<forall>x \<in> valid_ENC n ` FOV \<phi>. wf n x" by auto
  { fix r w I assume "r \<in> FOV \<phi>" and *: "length I = n" "wf_interp_for_formula (w, I) \<phi>"
    then obtain p where p: "I ! r = Inl p" by (cases "I ! r") auto
    moreover from `r \<in> FOV \<phi>` assms *(1) have "r < length I" by (auto dest!: max_idx_vars)
    ultimately have "Inl p \<in> set I" by (auto simp add: in_set_conv_nth)
    with *(2) have "p < length w" by auto
    with *(2) obtain a where a: "w ! p = a" "a \<in> set \<Sigma>" by auto
    from `r < length I` p *(1) `a \<in> set \<Sigma>`
      have "[enc_atom I p a] \<in> lang n (arbitrary_except n [(r, True)] \<Sigma>)"
      by (intro enc_atom_lang_arbitrary_except_True[OF _ _ _  _ subset_refl]) auto
    moreover
    from `r < length I` p *(1) `a \<in> set \<Sigma>` *(2)  `p < length w`
      have "take p (enc (w, I)) \<in> star (lang n (arbitrary_except n [(r, False)] \<Sigma>))"
      by (auto simp: in_set_conv_nth intro!: Ball_starI enc_atom_lang_arbitrary_except_False) auto  
    moreover
    from `r < length I` p *(1) `a \<in> set \<Sigma>` *(2)  `p < length w`
      have "drop (Suc p) (enc (w, I)) \<in> star (lang n (arbitrary_except n [(r, False)] \<Sigma>))"
      by (auto simp: in_set_conv_nth intro!: Ball_starI enc_atom_lang_arbitrary_except_False) auto
    ultimately have "take p (enc (w, I)) @ [enc_atom I p a] @ drop (p + 1) (enc (w, I)) \<in>
      lang n (valid_ENC n r)" using `0 < n` unfolding valid_ENC_def by (auto simp del: append.simps)
    with `p < length w` a have "enc (w, I) \<in> lang n (valid_ENC n r)"
      using id_take_nth_drop[of p "enc (w, I)"] by auto
  }
  hence "?R \<subseteq> ?L" using lang_flatten_INTERSECT[OF finite nonempty wf_rexp] by (auto simp add: ENC_def)
  moreover
  { fix x assume "x \<in> (\<Inter>r \<in> valid_ENC n ` FOV \<phi>. lang n r)"
    hence r: "\<forall>r \<in> FOV \<phi>. x \<in> lang n (valid_ENC n r)" by blast
    have "length (dec_interp n (FOV \<phi>) x) = n" unfolding dec_interp_def by simp
    moreover
    { fix r assume "r \<in> FOV \<phi>"
      with assms have "r < n" using max_idx_vars[of n \<phi>] by auto
      from `r \<in> FOV \<phi>` r obtain u v w where uvw: "x = u @ v @ w"
        "u \<in> star (lang n (arbitrary_except n [(r, False)] \<Sigma>))"
        "v \<in> lang n (arbitrary_except n [(r, True)] \<Sigma>)"
        "w \<in> star (lang n (arbitrary_except n [(r, False)] \<Sigma>))" using `0 < n` unfolding valid_ENC_def by fastforce
      hence "length u < length x" "\<And>i. i < length x \<longrightarrow> snd (x ! i) ! r \<longleftrightarrow> i = length u"
         by (auto simp: nth_append nth_Cons' split: split_if_asm
            dest!: arbitrary_except[OF _ `r < n`]
            dest: star_arbitrary_except[OF _ `r < n`, of u]
            elim!: iffD1[OF star_arbitrary_except[OF _ `r < n`, of w False]]) auto
      hence "\<exists>!p. p < length x \<and> snd (x ! p) ! r" by auto
    } note * = this
    have x_wf_word: "wf_word n x" using wf_lang_wf_word[OF wf_rexp_valid_ENC] False r by auto
    with * have "x = enc (dec_word x, dec_interp n (FOV \<phi>) x)" by (intro sym[OF enc_dec]) auto
    moreover
    from * have "wf_interp_for_formula (dec_word x, dec_interp n (FOV \<phi>) x) \<phi>"
      using r False x_wf_word[unfolded wf_word, unfolded \<sigma>_def] assms
      apply (auto simp: dec_word_def split: sum.splits)
      apply fastforce
      using Inl_dec_interp_length[OF ballI] apply blast
      using Inr_dec_interp_length apply blast
      using dec_interp_Inl[OF _ bspec[OF max_idx_vars], of _ "FOV \<phi>" n \<phi> x] apply force
      using dec_interp_Inr[OF _ bspec[OF max_idx_vars], of _ "FOV \<phi>" n \<phi> x] apply fastforce
      done
    ultimately have "x \<in> {enc (w, I) | w I. length I = n \<and> wf_interp_for_formula (w, I) \<phi>}" by blast
  }
  with False lang_flatten_INTERSECT[OF finite nonempty wf_rexp] have "?L \<subseteq> ?R" by (auto simp add: ENC_def)
  ultimately show ?thesis by blast
qed

subsection {* Welldefinedness of enc wrt. Models *}

lemma enc_alt_def:
 "enc (w, x # I) = map_index (\<lambda>n (a, bs). (a, (case x of Inl p \<Rightarrow> n = p | Inr P \<Rightarrow> n \<in> P) # bs)) (enc (w, I))"
by (auto simp: comp_def)

lemma enc_extend_interp: "enc (w, I) = enc (w', I') \<Longrightarrow> enc (w, x # I) = enc (w', x # I')"
unfolding enc_alt_def by auto

lemma wf_interp_for_formula_Exists: 
 "\<lbrakk>wf_formula (length I) (Exists \<phi>); w \<noteq> []\<rbrakk>\<Longrightarrow>
  wf_interp_for_formula (w, I) (Exists \<phi>) \<longleftrightarrow> (\<forall>p < length w. wf_interp_for_formula (w, Inl p # I) \<phi>)"
apply (clarsimp split: sum.splits split_if_asm)
apply safe
apply (metis (hide_lams) DiffI Suc_pred gr0I nth_Cons_0 nth_Cons_Suc singleton_iff sum.simps(4))
apply (metis diff_Suc_Suc diff_zero nat.exhaust nth_Cons_Suc)
apply (metis length_greater_0_conv)
apply (metis length_greater_0_conv)
apply (metis length_greater_0_conv)
apply (metis diff_Suc_Suc gr0_implies_Suc length_greater_0_conv minus_nat.diff_0 nth_Cons_Suc)
apply (auto simp: nth_Cons' split: split_if_asm)
done

lemma wf_interp_for_formula_any_Inl: "wf_interp_for_formula (w, Inl p # I) \<phi> \<Longrightarrow>
  \<forall>p < length w. wf_interp_for_formula (w, Inl p # I) \<phi>"
by (auto simp: nth_Cons' split: split_if_asm)

lemma wf_interp_for_formula_EXISTS: 
 "\<lbrakk>wf_formula (length I) (EXISTS \<phi>); w \<noteq> []\<rbrakk>\<Longrightarrow>
  wf_interp_for_formula (w, I) (EXISTS \<phi>) \<longleftrightarrow> (\<forall>P \<subseteq> {0 .. length w - 1}. wf_interp_for_formula (w, Inr P # I) \<phi>)"
apply (clarsimp split: sum.splits split_if_asm)
apply safe
apply (cases w)
apply auto [2]
apply (metis Suc_pred gr0I nth_Cons_Suc)
apply (metis (hide_lams) DiffI Suc_pred gr0I nth_Cons_0 nth_Cons_Suc singleton_iff sum.simps(4))
unfolding sym[OF length_greater_0_conv] nth_Cons' One_nat_def
apply auto [2]
apply (metis empty_subsetI)
apply (metis empty_subsetI)
apply (metis empty_subsetI neq0_conv)
done

lemma wf_interp_for_formula_any_Inr: "wf_interp_for_formula (w, Inr P # I) \<phi> \<Longrightarrow>
  \<forall>P \<subseteq> {0 .. length w - 1}. wf_interp_for_formula (w, Inr P # I) \<phi>"
by (cases w) (auto simp: nth_Cons' split: sum.splits split_if_asm)

lemma enc_word_length: "enc (w, I) = enc (w', I') \<Longrightarrow> length w = length w'"
by (auto elim: map_index_eq_imp_length_eq)

lemma enc_length: 
  assumes "w \<noteq> []" "enc (w, I) = enc (w', I')"
  shows "length I = length I'"
using assms
  length_map[of "(\<lambda>x. case x of Inl p \<Rightarrow> 0 = p | Inr P \<Rightarrow> 0 \<in> P)" I]
  length_map[of "(\<lambda>x. case x of Inl p \<Rightarrow> 0 = p | Inr P \<Rightarrow> 0 \<in> P)" I']
by (induct rule: list_induct2[OF enc_word_length[OF assms(2)]]) auto

lemma wf_interp_for_formula_Or:
  "\<lbrakk>wf_formula n (Or \<phi>1 \<phi>2); length I = n\<rbrakk> \<Longrightarrow> wf_interp_for_formula (w, I) (Or \<phi>1 \<phi>2) =
     (wf_interp_for_formula (w, I) \<phi>1 \<and> wf_interp_for_formula (w, I) \<phi>2)"
by auto

lemma wf_interp_for_formula_And:
  "\<lbrakk>wf_formula n (And \<phi>1 \<phi>2); length I = n\<rbrakk> \<Longrightarrow> wf_interp_for_formula (w, I) (And \<phi>1 \<phi>2) =
     (wf_interp_for_formula (w, I) \<phi>1 \<and> wf_interp_for_formula (w, I) \<phi>2)"
by auto

lemma max_idx_Suc: "pre_wf_formula n \<phi> \<Longrightarrow> pre_wf_formula (Suc n) \<phi>"
by (induct \<phi> arbitrary: n) auto

lemma wf_formula_Suc: "wf_formula n \<phi> \<Longrightarrow> wf_formula (Suc n) \<phi>"
by (auto simp: max_idx_Suc)

lemma enc_wf_interp: 
 assumes "wf_formula (length I) \<phi>" "wf_interp_for_formula (w, I) \<phi>"
 shows "wf_interp_for_formula (dec_word (enc (w, I)), dec_interp (length I) (FOV \<phi>) (enc (w, I))) \<phi>"
 (is "wf_interp_for_formula (_, ?dec) \<phi>")
unfolding dec_word_enc
proof -
  { fix i assume i: "i \<in> FOV \<phi>"
    with assms(2) have "\<exists>p. I ! i = Inl p" by (cases "I ! i") auto
    with i assms have "\<exists>!p. p < length (enc (w, I)) \<and> snd (enc (w, I) ! p) ! i"
      by (intro enc_unique[of w I i]) (auto intro!: bspec[OF max_idx_vars] split: sum.splits)
  } note * = this
  have "\<forall>x \<in> set ?dec. sum_case (\<lambda>p. p < length w) (\<lambda>P. \<forall>p\<in>P. p < length w) x"
  proof (intro ballI, split sum.split, safe)
    fix p assume "Inl p \<in> set ?dec"
    thus "p < length w" using Inl_dec_interp_length[OF ballI[OF *]] by auto
  next
    fix p P assume "Inr P \<in> set ?dec" "p \<in> P"
    thus "p < length w" using Inr_dec_interp_length by fastforce
  qed
  thus "wf_interp_for_formula (w, ?dec) \<phi>"
  using assms
    dec_interp_Inl[of _ "FOV \<phi>" "length I" "enc (w, I)", OF _ bspec[OF max_idx_vars]]
    dec_interp_Inr[of _ "FOV \<phi>" "length I" "enc (w, I)", OF _ bspec[OF max_idx_vars]]
  by (fastforce split: sum.splits)
qed

lemma enc_welldef: "\<lbrakk>enc (w, I) = enc (w', I'); wf_formula (length I) \<phi>;
  wf_interp_for_formula (w, I) \<phi>; wf_interp_for_formula (w', I') \<phi>\<rbrakk> \<Longrightarrow>
  satisfies (w, I) \<phi> \<longleftrightarrow> satisfies (w', I') \<phi>"
proof (induction \<phi> arbitrary: I I')
  case (Q a m)
  let ?dec = "\<lambda>w I. (dec_word (enc (w, I)), dec_interp (length I) (FOV (Q a m)) (enc (w, I)))"
  from Q(2,3) have "satisfies (w, I) (Q a m) = satisfies (?dec w I) (Q a m)"
    unfolding dec_word_enc 
    using dec_interp_enc_Inl[of "length I" "{m}" w I m]
    by (auto intro: nth_mem dest: dec_interp_not_Inr split: sum.splits) (metis nth_mem)+
  moreover
  from Q(3) have *: "w \<noteq> []" by simp
  from Q(2,4) have "satisfies (w', I') (Q a m) = satisfies (?dec w' I') (Q a m)"
    unfolding dec_word_enc enc_length[OF * Q(1)]
    using dec_interp_enc_Inl[of "length I'" "{m}" w' I' m]
    by (auto dest: dec_interp_not_Inr split: sum.splits) (metis nth_mem)+
  ultimately show ?case unfolding Q(1) enc_length[OF * Q(1)] ..
next
  case (Less m n)
  let ?dec = "\<lambda>w I. (dec_word (enc (w, I)), dec_interp (length I) (FOV (Less m n)) (enc (w, I)))"
  from Less(2,3) have "satisfies (w, I) (Less m n) = satisfies (?dec w I) (Less m n)"
    unfolding dec_word_enc
    using dec_interp_enc_Inl[of "length I" "{m, n}" w I m] dec_interp_enc_Inl[of "length I" "{m, n}" w I n]
    by (fastforce intro: nth_mem dest: dec_interp_not_Inr split: sum.splits)
  moreover
  from Less(3) have *: "w \<noteq> []" by simp
  from Less(2,4) have "satisfies (w', I') (Less m n) = satisfies (?dec w' I') (Less m n)"
    unfolding dec_word_enc enc_length[OF * Less(1)]
    using dec_interp_enc_Inl[of "length I'" "{m, n}" w' I' m] dec_interp_enc_Inl[of "length I'" "{m, n}" w' I' n]
    by (fastforce intro: nth_mem dest: dec_interp_not_Inr split: sum.splits)
  ultimately show ?case unfolding Less(1) enc_length[OF * Less(1)] ..
next
  case (In m M)
  let ?dec = "\<lambda>w I. (dec_word (enc (w, I)), dec_interp (length I) (FOV (In m M)) (enc (w, I)))"
  from In(2,3) have "satisfies (w, I) (In m M) = satisfies (?dec w I) (In m M)"
    unfolding dec_word_enc
    using dec_interp_enc_Inl[of "length I" "{m}" w I m] dec_interp_enc_Inr[of "length I" "{m}" w I M]
    by (auto dest: dec_interp_not_Inr dec_interp_not_Inl split: sum.splits) (metis nth_mem)+
  moreover
  from In(3) have *: "w \<noteq> []" by simp
  from In(2,4) have "satisfies (w', I') (In m M) = satisfies (?dec w' I') (In m M)"
    unfolding dec_word_enc enc_length[OF * In(1)]
    using dec_interp_enc_Inl[of "length I'" "{m}" w' I' m] dec_interp_enc_Inr[of "length I'" "{m}" w' I' M]
    by (auto dest: dec_interp_not_Inr dec_interp_not_Inl split: sum.splits) (metis nth_mem)+
  ultimately show ?case unfolding In(1) enc_length[OF * In(1)] ..
next
  case (Or \<phi>1 \<phi>2) show ?case unfolding satisfies.simps(5)
  proof (intro disj_cong)
    from Or(3-6) show "satisfies (w, I) \<phi>1 = satisfies (w', I') \<phi>1"
      by (intro Or(1)) auto
  next
    from Or(3-6) show "satisfies (w, I) \<phi>2 = satisfies (w', I') \<phi>2"
      by (intro Or(2)) auto
  qed
next
  case (And \<phi>1 \<phi>2) show ?case unfolding satisfies.simps(6)
  proof (intro conj_cong)
    from And(3-6) show "satisfies (w, I) \<phi>1 = satisfies (w', I') \<phi>1"
      by (intro And(1)) auto
  next
    from And(3-6) show "satisfies (w, I) \<phi>2 = satisfies (w', I') \<phi>2"
      by (intro And(2)) auto
  qed
next
  case (Exists \<phi>)
  hence "w \<noteq> []" "w' \<noteq> []" by auto
  hence length: "length w = length w'" "length I = length I'"
    using enc_word_length[OF Exists.prems(1)] enc_length[OF _ Exists.prems(1)] by auto
  show ?case
  proof
    assume "satisfies (w, I) (Exists \<phi>)"
    with Exists.prems(3) obtain p where "p < length w" "satisfies (w, Inl p # I) \<phi>"
      by (auto intro: ord_less_eq_trans[OF le_imp_less_Suc Suc_pred])
    moreover
    with Exists.prems have "satisfies (w', Inl p # I') \<phi>"
      apply (intro iffD1[OF Exists.IH[of "Inl p # I" "Inl p # I'"]])
      apply (elim enc_extend_interp)
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_Exists[OF _ `w \<noteq> []`])
      apply (blast dest!: wf_interp_for_formula_Exists[OF _ `w' \<noteq> []`, of I', unfolded length[symmetric]])
      apply assumption
      done
    ultimately show "satisfies (w', I') (Exists \<phi>)" using length by (auto intro!: exI[of _ p])
  next
    assume "satisfies (w', I') (Exists \<phi>)"
    with Exists.prems(1,2,4) obtain p where "p < length w'" "satisfies (w', Inl p # I') \<phi>"
      by (auto intro: ord_less_eq_trans[OF le_imp_less_Suc Suc_pred])
    moreover
    with Exists.prems have "satisfies (w, Inl p # I) \<phi>"
      apply (intro iffD2[OF Exists.IH[of "Inl p # I" "Inl p # I'"]])
      apply (elim enc_extend_interp)
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_Exists[OF _ `w \<noteq> []`, of I, unfolded length(1)])
      apply (blast dest!: wf_interp_for_formula_Exists[OF _ `w' \<noteq> []`, of I', unfolded length(2)[symmetric]])
      apply assumption
      done
    ultimately show "satisfies (w, I) (Exists \<phi>)" using length by (auto intro!: exI[of _ p])
  qed
next
  case (EXISTS \<phi>)
  hence "w \<noteq> []" "w' \<noteq> []" by auto
  hence length: "length w = length w'" "length I = length I'"
    using enc_word_length[OF EXISTS.prems(1)] enc_length[OF _ EXISTS.prems(1)] by auto
  show ?case
  proof
    assume "satisfies (w, I) (EXISTS \<phi>)"
    then obtain P where "P \<subseteq> {0 .. length w - 1}" "satisfies (w, Inr P # I) \<phi>" by auto
    moreover
    with EXISTS.prems have "satisfies (w', Inr P # I') \<phi>"
      apply (intro iffD1[OF EXISTS.IH[of "Inr P # I" "Inr P # I'"]])
      apply (elim enc_extend_interp)
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_EXISTS[OF _ `w \<noteq> []`])
      apply (blast dest!: wf_interp_for_formula_EXISTS[OF _ `w' \<noteq> []`, of I', unfolded length[symmetric]])
      apply assumption
      done
    ultimately show "satisfies (w', I') (EXISTS \<phi>)" using length by (auto intro!: exI[of _ P])
  next
    assume "satisfies (w', I') (EXISTS \<phi>)"
    then obtain P where P: "P \<subseteq> {0 .. length w' - 1}" "satisfies (w', Inr P # I') \<phi>" by auto
    moreover
    with EXISTS.prems have "satisfies (w, Inr P # I) \<phi>"
      apply (intro iffD2[OF EXISTS.IH[of "Inr P # I" "Inr P # I'"]])
      apply (elim enc_extend_interp)
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_EXISTS[OF _ `w \<noteq> []`, of I, unfolded length(1)])
      apply (blast dest!: wf_interp_for_formula_EXISTS[OF _ `w' \<noteq> []`, of I', unfolded length(2)[symmetric]])
      apply assumption
      done
    ultimately show "satisfies (w, I) (EXISTS \<phi>)" using length by (auto intro!: exI[of _ P])
  qed
qed auto

lemma lang\<^isub>M\<^isub>S\<^isub>O_Or:
  assumes "wf_formula n (Or \<phi>\<^isub>1 \<phi>\<^isub>2)"
  shows "lang\<^isub>M\<^isub>S\<^isub>O n (Or \<phi>\<^isub>1 \<phi>\<^isub>2) \<subseteq>
    (lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1 \<union> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2) \<inter> {enc (w, I) | w I. length I = n \<and> wf_interp_for_formula (w, I) (Or \<phi>\<^isub>1 \<phi>\<^isub>2)}"
    (is "_ \<subseteq> (?L1 \<union> ?L2) \<inter> ?ENC")
proof (intro equalityI subsetI)
  fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Or \<phi>\<^isub>1 \<phi>\<^isub>2)"
  then obtain w I where
    *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Or \<phi>\<^isub>1 \<phi>\<^isub>2)" "length I = n" "length w > 0" and
     "satisfies (w, I) \<phi>\<^isub>1 \<or> satisfies (w, I) \<phi>\<^isub>2" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
  thus "x \<in> (?L1 \<union> ?L2) \<inter> ?ENC"
  proof (elim disjE)
    assume "satisfies (w, I) \<phi>\<^isub>1"
    with * have "x \<in> ?L1" using assms unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by (fastforce)
    with * show ?thesis by auto
  next
    assume "satisfies (w, I) \<phi>\<^isub>2"
    with * have "x \<in>?L2" using assms unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by (fastforce)
    with * show ?thesis by auto
  qed
qed

lemma lang\<^isub>M\<^isub>S\<^isub>O_And:
  assumes "wf_formula n (And \<phi>\<^isub>1 \<phi>\<^isub>2)"
  shows "lang\<^isub>M\<^isub>S\<^isub>O n (And \<phi>\<^isub>1 \<phi>\<^isub>2) \<subseteq>
    lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1 \<inter> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2 \<inter> {enc (w, I) | w I. length I = n \<and> wf_interp_for_formula (w, I) (And \<phi>\<^isub>1 \<phi>\<^isub>2)}"
    (is "_ \<subseteq> ?L1 \<inter> ?L2 \<inter> ?ENC")
using assms unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by (fastforce)

subsection {* From MSO to Regular expressions *}

fun rexp_of :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "rexp_of n (Q a m) =
    Inter (TIMES [rexp.Not Zero, arbitrary_except n [(m, True)] [a], rexp.Not Zero])
    (ENC n (Q a m))"
| "rexp_of n (Less m1 m2) = (if m1 = m2 then Zero else
    Inter (TIMES [rexp.Not Zero, arbitrary_except n [(m1, True)] \<Sigma>,
      rexp.Not Zero, arbitrary_except n [(m2, True)] \<Sigma>,
      rexp.Not Zero]) (ENC n (Less m1 m2)))"
| "rexp_of n (In m M) = 
    Inter (TIMES [rexp.Not Zero, arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>, rexp.Not Zero])
    (ENC n (In m M))"
| "rexp_of n (Not \<phi>) = Inter (rexp.Not (rexp_of n \<phi>)) (ENC n (Not \<phi>))"
| "rexp_of n (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = Inter (Plus (rexp_of n \<phi>\<^isub>1) (rexp_of n \<phi>\<^isub>2)) (ENC n (Or \<phi>\<^isub>1 \<phi>\<^isub>2))"
| "rexp_of n (And \<phi>\<^isub>1 \<phi>\<^isub>2) = INTERSECT [rexp_of n \<phi>\<^isub>1, rexp_of n \<phi>\<^isub>2, ENC n (And \<phi>\<^isub>1 \<phi>\<^isub>2)]"
| "rexp_of n (Exists \<phi>) = Pr (rexp_of (n + 1) \<phi>)"
| "rexp_of n (EXISTS \<phi>) = Pr (rexp_of (n + 1) \<phi>)"

fun rexp_of_alt :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "rexp_of_alt n (Q a m) =
    TIMES [rexp.Not Zero, arbitrary_except n [(m, True)] [a], rexp.Not Zero]"
| "rexp_of_alt n (Less m1 m2) = (if m1 = m2 then Zero else
    TIMES [rexp.Not Zero, arbitrary_except n [(m1, True)] \<Sigma>,
      rexp.Not Zero, arbitrary_except n [(m2, True)] \<Sigma>,
      rexp.Not Zero])"
| "rexp_of_alt n (In m M) = 
    TIMES [rexp.Not Zero, arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>, rexp.Not Zero]"
| "rexp_of_alt n (Not \<phi>) = rexp.Not (rexp_of_alt n \<phi>)"
| "rexp_of_alt n (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = Plus (rexp_of_alt n \<phi>\<^isub>1) (rexp_of_alt n \<phi>\<^isub>2)"
| "rexp_of_alt n (And \<phi>\<^isub>1 \<phi>\<^isub>2) = Inter (rexp_of_alt n \<phi>\<^isub>1) (rexp_of_alt n \<phi>\<^isub>2)"
| "rexp_of_alt n (Exists \<phi>) = Pr (Inter (rexp_of_alt (n + 1) \<phi>) (ENC (n + 1) \<phi>))"
| "rexp_of_alt n (EXISTS \<phi>) = Pr (Inter (rexp_of_alt (n + 1) \<phi>) (ENC (n + 1) \<phi>))"

definition "rexp_of' n \<phi> = Inter (rexp_of_alt n \<phi>) (ENC n \<phi>)"
(*
abbreviation except where
  "except m1 m2 \<equiv> if m1 < m2 then [(m1, True), (m2, False)] else [(m2, False), (m1, True)]"

fun rexp_of'' :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "rexp_of'' n (Q a m) =
    TIMES [Star (arbitrary_except n [(m, False)] \<Sigma>),
      arbitrary_except n [(m, True)] [a],
      Star (arbitrary_except n [(m, False)] \<Sigma>)]"
| "rexp_of'' n (Less m1 m2) = (if m1 = m2 then Zero else
    TIMES [Star (arbitrary_except n [(m1, False), (m2, False)] \<Sigma>),
      arbitrary_except n (except m1 m2) \<Sigma>,
      Star (arbitrary_except n [(m1, False), (m2, False)] \<Sigma>),
      arbitrary_except n (except m2 m1) \<Sigma>,
      Star (arbitrary_except n [(m1, False), (m2, False)] \<Sigma>)])"
| "rexp_of'' n (In m M) = 
    TIMES [Star (arbitrary_except n [(m, False)] \<Sigma>),
      arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>,
      Star (arbitrary_except n [(m, False)] \<Sigma>)]"
| "rexp_of'' n (Not \<phi>) = Inter (rexp.Not (rexp_of'' n \<phi>)) (ENC n (Not \<phi>))"
| "rexp_of'' n (Or \<phi>\<^isub>1 \<phi>\<^isub>2) = Inter (Plus (rexp_of'' n \<phi>\<^isub>1) (rexp_of'' n \<phi>\<^isub>2)) (ENC n (Or \<phi>\<^isub>1 \<phi>\<^isub>2))"
| "rexp_of'' n (And \<phi>\<^isub>1 \<phi>\<^isub>2) = INTERSECT [rexp_of'' n \<phi>\<^isub>1, rexp_of'' n \<phi>\<^isub>2, ENC n (And \<phi>\<^isub>1 \<phi>\<^isub>2)]"
| "rexp_of'' n (Exists \<phi>) = Pr (rexp_of'' (n + 1) \<phi>)"
| "rexp_of'' n (EXISTS \<phi>) = Pr (rexp_of'' (n + 1) \<phi>)"
*)
lemma length_dec_interp[simp]: "length (dec_interp n FO x) = n"
unfolding dec_interp_def by auto

theorem lang\<^isub>M\<^isub>S\<^isub>O_rexp_of: "wf_formula n \<phi> \<Longrightarrow> lang\<^isub>M\<^isub>S\<^isub>O n \<phi> = lang n (rexp_of n \<phi>) - {[]}"
   (is "_ \<Longrightarrow> _ = ?L n \<phi>")
proof (induct \<phi> arbitrary: n)
  case (Q a m)
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Q a m)"
    then obtain w I where
      *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Q a m)" "satisfies (w, I) (Q a m)"
       "length I = n"
      unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by blast
    with Q(1) obtain p where p: "p < length w" "I ! m = Inl p" "w ! p = a"
      by (auto simp: all_set_conv_all_nth split: sum.splits)
    with *(1) have "x = take p (enc (w, I)) @ [enc_atom I p a] @ drop (p + 1) (enc (w, I))"
      using id_take_nth_drop[of p "enc (w, I)"] by auto
    moreover from *(4) Q(1) p(2)
     have "[enc_atom I p a] \<in> lang n (arbitrary_except n [(m, True)] [a])"
       by (intro enc_atom_lang_arbitrary_except_True) auto
    moreover from *(2,4) have "take p (enc (w, I)) \<in> lang n (rexp.Not Zero)"
      by (auto intro!: enc_atom_\<sigma> dest!: in_set_takeD)
    moreover from *(2,4) have "drop (Suc p) (enc (w, I)) \<in> lang n (rexp.Not Zero)"
      by (auto intro!: enc_atom_\<sigma> dest!: in_set_dropD)
    ultimately show "x \<in> ?L n (Q a m)" using *(1,2,4)
      unfolding rexp_of.simps lang.simps(5,8) rexp_of_list.simps Int_Diff lang_ENC[OF Q]
      by (auto elim: ssubst simp del: o_apply append.simps lang.simps)
  next
    fix x assume x: "x \<in> ?L n (Q a m)"
    with Q obtain w I p where m: "I ! m = Inl p" "m < length I" and
      wI: "x = enc (w, I)" "length I = n" "wf_interp_for_formula (w, I) (Q a m)"
      unfolding rexp_of.simps lang.simps lang_ENC[OF Q] Int_Diff by atomize_elim (auto split: sum.splits)
    hence "wf_interp_for_formula (dec_word x, dec_interp n {m} x) (Q a m)" unfolding wI(1)
      using enc_wf_interp[OF Q(1)[folded wI(2)]] by auto
    moreover
    from x obtain u1 u u2 where "x = u1 @ u @ u2" "u \<in> lang n (arbitrary_except n [(m, True)] [a])"
      unfolding rexp_of.simps lang.simps rexp_of_list.simps using concE by fast
    with Q(1) obtain v where v: "x = u1 @ [v] @ u2" "snd v ! m" "fst v = a"
      using arbitrary_except[of u n m True "[a]"] by fastforce
    hence u: "length u1 < length x" by auto
    { from v have "snd (x ! length u1) ! m" by auto
      moreover
      from m wI have "p < length x" "snd (x ! p) ! m"
        by (fastforce intro: nth_mem split: sum.splits)+
      moreover
      from m wI have ex1: "\<exists>!p. p < length x \<and> snd (x ! p) ! m" unfolding wI(1) by (intro enc_unique) auto
      ultimately have "p = length u1" using u by auto
    } note * = this
    from v have "v = enc (w, I) ! length u1" unfolding wI(1) by simp
    hence "a = w ! length u1" using nth_map[OF u, of fst] unfolding wI(1) v(3)[symmetric] by auto
    with * m wI have "satisfies (dec_word x, dec_interp n {m} x) (Q a m)"
      unfolding dec_word_enc[of w I, folded wI(1)]
      by (auto simp del: enc.simps dest: dec_interp_not_Inr split: sum.splits)
         (fastforce dest!: dec_interp_enc_Inl intro: nth_mem split: sum.splits)
    moreover from wI have "wf_word n x" unfolding wf_word by (auto intro!: enc_atom_\<sigma>)
    ultimately show "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Q a m)" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def using m wI(3)
      by (auto simp del: enc.simps intro!: exI[of _ "dec_word x"] exI[of _ "dec_interp n {m} x"]
        intro: sym[OF enc_dec[OF _ ballI[OF impI[OF enc_unique[of w I, folded wI(1)]]]]])
  qed
next
  case (Less m m')
  show ?case
  proof (cases "m = m'")
    case False
    thus ?thesis
    proof (intro equalityI subsetI)
      fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Less m m')"
      then obtain w I where
        *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Less m m')" "satisfies (w, I) (Less m m')"
         "length I = n"
        unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by blast
      with Less(1) obtain p q where pq: "p < length w" "I ! m = Inl p" "q < length w" "I ! m' = Inl q" "p < q"
        by (auto simp: all_set_conv_all_nth split: sum.splits)
      with *(1) have "x = take p (enc (w, I)) @ [enc_atom I p (w ! p)] @ drop (p + 1) (enc (w, I))"
        using id_take_nth_drop[of p "enc (w, I)"] by auto
      also have "drop (p + 1) (enc (w, I)) = take (q - p - 1) (drop (p + 1) (enc (w, I))) @
        [enc_atom I q (w ! q)] @ drop (q - p) (drop (p + 1) (enc (w, I)))" (is "_ = ?LHS")
        using id_take_nth_drop[of "q - p - 1" "drop (p + 1) (enc (w, I))"] pq by auto
      finally have "x = take p (enc (w, I)) @ [enc_atom I p (w ! p)] @ ?LHS" .
      moreover from *(2,4) Less(1) pq(1,2)
       have "[enc_atom I p (w ! p)] \<in> lang n (arbitrary_except n [(m, True)] \<Sigma>)"
         by (intro enc_atom_lang_arbitrary_except_True) auto
      moreover from *(2,4) Less(1) pq(3,4)
       have "[enc_atom I q (w ! q)] \<in> lang n (arbitrary_except n [(m', True)] \<Sigma>)"
         by (intro enc_atom_lang_arbitrary_except_True) auto
      moreover from *(2,4) have "take p (enc (w, I)) \<in> lang n (rexp.Not Zero)"
        by (auto intro!: enc_atom_\<sigma> dest!: in_set_takeD)
      moreover from *(2,4) have "take (q - p - 1) (drop (Suc p) (enc (w, I))) \<in> lang n (rexp.Not Zero)"
        by (auto intro!: enc_atom_\<sigma> dest!: in_set_dropD in_set_takeD)
      moreover from *(2,4) have "drop (q - p) (drop (Suc p) (enc (w, I))) \<in> lang n (rexp.Not Zero)"
        by (auto intro!: enc_atom_\<sigma> dest!: in_set_dropD)
      ultimately show "x \<in> ?L n (Less m m')" using *(1,2,4)
        unfolding rexp_of.simps lang.simps(5,8) rexp_of_list.simps Int_Diff lang_ENC[OF Less] if_not_P[OF False]
        by (auto elim: ssubst simp del: o_apply append.simps lang.simps)
    next
      fix x assume x: "x \<in> ?L n (Less m m')"
      with Less obtain w I where
        wI: "x = enc (w, I)" "length I = n" "wf_interp_for_formula (w, I) (Less m m')"
        unfolding rexp_of.simps lang.simps lang_ENC[OF Less] Int_Diff if_not_P[OF False]
        by (fastforce split: sum.splits)
      with Less obtain p p' where m: "I ! m = Inl p" "m < length I" "I ! m' = Inl p'" "m' < length I"
        by (auto split: sum.splits)
      with wI have "wf_interp_for_formula (dec_word x, dec_interp n {m, m'} x) (Less m m')" unfolding wI(1)
        using enc_wf_interp[OF Less(1)[folded wI(2)]] by auto
      moreover
      from x obtain u1 u u2 u' u3 where "x = u1 @ u @ u2 @ u' @ u3"
        "u \<in> lang n (arbitrary_except n [(m, True)] \<Sigma>)" 
        "u' \<in> lang n (arbitrary_except n [(m', True)] \<Sigma>)"
        unfolding rexp_of.simps lang.simps rexp_of_list.simps if_not_P[OF False] using concE by fast
      with Less(1) obtain v v' where v: "x = u1 @ [v] @ u2 @ [v'] @ u3" "snd v ! m" "snd v' ! m'"
        using arbitrary_except[of u n m True \<Sigma>] arbitrary_except[of u' n m' True \<Sigma>] by fastforce
      hence u: "length u1 < length x" and u': "Suc (length u1 + length u2) < length x" (is "?u' < _") by auto
      { from v have "snd (x ! length u1) ! m" by auto
        moreover
        from m wI have "p < length x" "snd (x ! p) ! m"
          by (fastforce intro: nth_mem split: sum.splits)+
        moreover
        from m wI have ex1: "\<exists>!p. p < length x \<and> snd (x ! p) ! m" unfolding wI(1) by (intro enc_unique) auto
        ultimately have "p = length u1" using u by auto
      }
      { from v have "snd (x ! ?u') ! m'" by (auto simp: nth_append)
        moreover
        from m wI have "p' < length x" "snd (x ! p') ! m'"
          by (fastforce intro: nth_mem split: sum.splits)+
        moreover
        from m wI have ex1: "\<exists>!p. p < length x \<and> snd (x ! p) ! m'" unfolding wI(1) by (intro enc_unique) auto
        ultimately have "p' = ?u'" using u' by auto
      } note * = this `p = length u1`
      with * m wI have "satisfies (dec_word x, dec_interp n {m, m'} x) (Less m m')"
        unfolding dec_word_enc[of w I, folded wI(1)]
        by (auto simp del: enc.simps dest: dec_interp_not_Inr split: sum.splits)
           (fastforce dest!: dec_interp_enc_Inl intro: nth_mem split: sum.splits)
      moreover from wI have "wf_word n x" unfolding wf_word by (auto intro!: enc_atom_\<sigma>)
      ultimately show "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Less m m')" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def using m wI(3)
        by (auto simp del: enc.simps intro!: exI[of _ "dec_word x"] exI[of _ "dec_interp n {m, m'} x"]
          intro: sym[OF enc_dec[OF _ ballI[OF impI[OF enc_unique[of w I, folded wI(1)]]]]])
    qed
  qed (simp add: lang\<^isub>M\<^isub>S\<^isub>O_def del: o_apply)
next
  case (In m M)
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (In m M)"
    then obtain w I where
      *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (In m M)" "satisfies (w, I) (In m M)"
       "length I = n"
      unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by blast
    with In(1) obtain p P where p: "p < length w" "I ! m = Inl p" "I ! M = Inr P" "p \<in> P"
      by (auto simp: all_set_conv_all_nth split: sum.splits)
    with *(1) have "x = take p (enc (w, I)) @ [enc_atom I p (w ! p)] @ drop (p + 1) (enc (w, I))"
      using id_take_nth_drop[of p "enc (w, I)"] by auto
    moreover
     have "[enc_atom I p (w ! p)] \<in> lang n (arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>)"
     proof (cases "m < M")
       case True with *(2,4) In(1) p show ?thesis
         by (intro enc_atom_lang_arbitrary_except_True2) (auto simp: min_absorb1 max_absorb2)
     next
       case False with *(2,4) In(1) p show ?thesis
         by (intro enc_atom_lang_arbitrary_except_True2) (auto simp: min_absorb2 max_absorb1)
     qed
    moreover from *(2,4) have "take p (enc (w, I)) \<in> lang n (rexp.Not Zero)"
      by (auto intro!: enc_atom_\<sigma> dest!: in_set_takeD)
    moreover from *(2,4) have "drop (Suc p) (enc (w, I)) \<in> lang n (rexp.Not Zero)"
      by (auto intro!: enc_atom_\<sigma> dest!: in_set_dropD)
    ultimately show "x \<in> ?L n (In m M)" using *(1,2,4)
      unfolding rexp_of.simps lang.simps(5,8) rexp_of_list.simps Int_Diff lang_ENC[OF In]
      by (auto elim: ssubst simp del: o_apply append.simps lang.simps)
  next
    fix x assume x: "x \<in> ?L n (In m M)"
    with In obtain w I where  wI: "x = enc (w, I)" "length I = n" "wf_interp_for_formula (w, I) (In m M)"
      unfolding rexp_of.simps lang.simps lang_ENC[OF In] Int_Diff by (fastforce split: sum.splits)
    with In obtain p P where m: "I ! m = Inl p" "m < length I" "I ! M = Inr P" "M < length I" by (auto split: sum.splits)
    with wI have "wf_interp_for_formula (dec_word x, dec_interp n {m} x) (In m M)" unfolding wI(1)
      using enc_wf_interp[OF In(1)[folded wI(2)]] by auto
    moreover
    from x obtain u1 u u2 where "x = u1 @ u @ u2"
      "u \<in> lang n (arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>)"
      unfolding rexp_of.simps lang.simps rexp_of_list.simps using concE by fast
    with In(1) obtain v where v: "x = u1 @ [v] @ u2" and "snd v ! min m M" "snd v ! max m M"
      using arbitrary_except2[of u n "min m M" True "max m M" True \<Sigma>] by fastforce
    hence v': "snd v ! m" "snd v ! M"
      by (induct "m < M") (auto simp: min_absorb1 min_absorb2 max_absorb1 max_absorb2)
    from v have u: "length u1 < length x" by auto
    { from v v' have "snd (x ! length u1) ! m" by auto
      moreover
      from m wI have "p < length x" "snd (x ! p) ! m"
        by (fastforce intro: nth_mem split: sum.splits)+
      moreover
      from m wI have ex1: "\<exists>!p. p < length x \<and> snd (x ! p) ! m" unfolding wI(1) by (intro enc_unique) auto
      ultimately have "p = length u1" using u by auto
    } note * = this
    from v v' have "v = enc (w, I) ! length u1" unfolding wI(1) by simp
    with v'(2) m(3,4) u wI(1) have "length u1 \<in> P" by auto
    with * m wI have "satisfies (dec_word x, dec_interp n {m} x) (In m M)"
      unfolding dec_word_enc[of w I, folded wI(1)]
      by (auto simp del: enc.simps dest: dec_interp_not_Inr dec_interp_not_Inl split: sum.splits)
         (auto simp del: enc.simps dest!: dec_interp_enc_Inl dec_interp_enc_Inr dest: nth_mem split: sum.splits)
    moreover from wI have "wf_word n x" unfolding wf_word by (auto intro!: enc_atom_\<sigma>)
    ultimately show "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (In m M)" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def using m wI(3)
      by (auto simp del: enc.simps intro!: exI[of _ "dec_word x"] exI[of _ "dec_interp n {m} x"]
        intro: sym[OF enc_dec[OF _ ballI[OF impI[OF enc_unique[of w I, folded wI(1)]]]]])
  qed
next
  case (Or \<phi>\<^isub>1 \<phi>\<^isub>2)
  from Or(3) have IH1: "lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1 = lang n (rexp_of n \<phi>\<^isub>1) - {[]}"
    by (intro Or(1)) auto
  from Or(3) have IH2: "lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2 = lang n (rexp_of n \<phi>\<^isub>2) - {[]}"
    by (intro Or(2)) auto
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Or \<phi>\<^isub>1 \<phi>\<^isub>2)" thus "x \<in> lang n (rexp_of n (Or \<phi>\<^isub>1 \<phi>\<^isub>2)) - {[]}"
      using lang\<^isub>M\<^isub>S\<^isub>O_Or[OF Or(3)] unfolding lang_ENC[OF Or(3)] rexp_of.simps lang.simps IH1 IH2 Int_Diff by auto
  next
    fix x assume "x \<in> lang n (rexp_of n (Or \<phi>\<^isub>1 \<phi>\<^isub>2)) - {[]}"
    then obtain w I where or: "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1 \<or> x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2" and wI: "x = enc (w, I)" "length I = n"
      "wf_interp_for_formula (w, I) (Or \<phi>\<^isub>1 \<phi>\<^isub>2)"
      unfolding lang_ENC[OF Or(3)] rexp_of.simps lang.simps IH1 IH2 Int_Diff by auto
    have "satisfies (w, I) \<phi>\<^isub>1 \<or> satisfies (w, I) \<phi>\<^isub>2"
    proof (intro mp[OF disj_mono[OF impI impI] or])
      assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1"
      with wI(2,3) Or(3) show "satisfies (w, I) \<phi>\<^isub>1"
        unfolding lang\<^isub>M\<^isub>S\<^isub>O_def wI(1) wf_interp_for_formula_Or[OF Or(3) wI(2)]
        by (auto simp del: enc.simps dest!: iffD2[OF enc_welldef[of _ _ _ _ \<phi>\<^isub>1]])
    next
      assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2"
      with wI(2,3) Or(3) show "satisfies (w, I) \<phi>\<^isub>2"
        unfolding lang\<^isub>M\<^isub>S\<^isub>O_def wI(1) wf_interp_for_formula_Or[OF Or(3) wI(2)]
        by (auto simp del: enc.simps dest!: iffD2[OF enc_welldef[of _ _ _ _ \<phi>\<^isub>2]])
    qed
    with wI show "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Or \<phi>\<^isub>1 \<phi>\<^isub>2)" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
  qed
next
  case (And \<phi>\<^isub>1 \<phi>\<^isub>2)
  from And(3) have IH1: "lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1 = lang n (rexp_of n \<phi>\<^isub>1) - {[]}"
    by (intro And(1)) auto
  from And(3) have IH2: "lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2 = lang n (rexp_of n \<phi>\<^isub>2) - {[]}"
    by (intro And(2)) auto
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (And \<phi>\<^isub>1 \<phi>\<^isub>2)" thus "x \<in> lang n (rexp_of n (And \<phi>\<^isub>1 \<phi>\<^isub>2)) - {[]}"
      using lang\<^isub>M\<^isub>S\<^isub>O_And[OF And(3)]
      unfolding lang_ENC[OF And(3)] rexp_of.simps rexp_of_list.simps lang.simps IH1 IH2 Int_Diff by auto
  next
    fix x assume "x \<in> lang n (rexp_of n (And \<phi>\<^isub>1 \<phi>\<^isub>2)) - {[]}"
    then obtain w I where "and": "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1 \<and> x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2" and wI: "x = enc (w, I)" "length I = n"
      "wf_interp_for_formula (w, I) (And \<phi>\<^isub>1 \<phi>\<^isub>2)"
      unfolding lang_ENC[OF And(3)] rexp_of.simps rexp_of_list.simps lang.simps IH1 IH2 Int_Diff by auto
    have "satisfies (w, I) \<phi>\<^isub>1 \<and> satisfies (w, I) \<phi>\<^isub>2"
    proof (intro mp[OF conj_mono[OF impI impI] "and"])
      assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>1"
      with wI(2,3) And(3) show "satisfies (w, I) \<phi>\<^isub>1"
        unfolding lang\<^isub>M\<^isub>S\<^isub>O_def wI(1) wf_interp_for_formula_And[OF And(3) wI(2)]
        by (auto simp del: enc.simps dest!: iffD2[OF enc_welldef[of _ _ _ _ \<phi>\<^isub>1]])
    next
      assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>\<^isub>2"
      with wI(2,3) And(3) show "satisfies (w, I) \<phi>\<^isub>2"
        unfolding lang\<^isub>M\<^isub>S\<^isub>O_def wI(1) wf_interp_for_formula_And[OF And(3) wI(2)]
        by (auto simp del: enc.simps dest!: iffD2[OF enc_welldef[of _ _ _ _ \<phi>\<^isub>2]])
    qed
    with wI show "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (And \<phi>\<^isub>1 \<phi>\<^isub>2)" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
  qed
next
  case (Not \<phi>)
  hence IH: "?L n \<phi> =  lang\<^isub>M\<^isub>S\<^isub>O n \<phi>" by simp
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Not \<phi>)"
    then obtain w I where
      *: "x = enc (w, I)" "wf_interp_for_formula (w, I) \<phi>" "length I = n" "length w > 0"
       and unsat: "\<not> (satisfies (w, I) \<phi>)"
       unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    { assume "x \<in> ?L n \<phi>"
      with IH have "satisfies (w, I) \<phi>" using enc_welldef[of _ _ w I \<phi>, OF _ _ _ *(2)] Not(2)
        unfolding *(1,3) lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    }
    with unsat have "x \<notin> ?L n \<phi>" by blast
    with * show "x \<in> ?L n (Not \<phi>)" unfolding rexp_of.simps lang.simps using lang_ENC[OF Not(2)]
      by (auto simp: comp_def intro!: enc_atom_\<sigma>)
  next
    fix x assume "x \<in> ?L n (Not \<phi>)"
    with IH have "x \<in> lang n (ENC n (Not \<phi>)) - {[]}" and x: "x \<notin> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>" by (auto simp del: o_apply)
    then obtain w I where *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Not \<phi>)" "length I = n"
      unfolding lang_ENC[OF Not(2)] by blast
    { assume "\<not> satisfies (w, I) (Not \<phi>)"
      with * have "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n \<phi>" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    }
    with x * show "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Not \<phi>)"  unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by blast
  qed
next
  case (Exists \<phi>)
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Exists \<phi>)"
    then obtain w I p where
      *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Exists \<phi>)"
       "length I = n" "length w > 0" "p \<in> {0 .. length w - 1}" "satisfies (w, Inl p # I) \<phi>"
       unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    with Exists(2) have "enc (w, Inl p # I) \<in> ?L (Suc n) \<phi>"
      by (intro subsetD[OF equalityD1[OF Exists(1)], of "Suc n" "enc (w, Inl p # I)"])
       (auto simp: lang\<^isub>M\<^isub>S\<^isub>O_def nth_Cons' ord_less_eq_trans[OF le_imp_less_Suc Suc_pred[OF *(4)]]
        split: split_if_asm sum.splits intro!: exI[of _ w] exI[of _ "Inl p # I"])
    with *(1) show "x \<in> ?L n (Exists \<phi>)"
      by (auto simp: map_index intro!: image_eqI[of _ "map \<pi>"] simp del: o_apply) (auto simp: \<pi>_def)
  next
    fix x assume "x \<in> ?L n (Exists \<phi>)"
    then obtain x' where x: "x = map \<pi> x'" and "x' \<in> ?L (Suc n) \<phi>" by (auto simp del: o_apply)
    with Exists(2) have "x' \<in> lang\<^isub>M\<^isub>S\<^isub>O (Suc n) \<phi>"
      by (intro subsetD[OF equalityD2[OF Exists(1)], of "Suc n" x'])
         (auto split: split_if_asm sum.splits)
    then obtain w I' where
      *: "x' = enc (w, I')" "wf_interp_for_formula (w, I') \<phi>" "length I' = Suc n" "satisfies (w, I') \<phi>"
       unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    moreover then obtain I\<^isub>0 I where "I' = I\<^isub>0 # I" by (cases I') auto
    moreover with Exists(2) *(2) obtain p where "I\<^isub>0 = Inl p" "p < length w"
      by (auto simp: nth_Cons' split: sum.splits split_if_asm)
    ultimately have "x = enc (w, I)" "wf_interp_for_formula (w, I) (Exists \<phi>)" "length I = n"
      "length w > 0" "satisfies (w, I) (Exists \<phi>)"using Exists(2) unfolding x
      by (auto simp: map_tl nth_Cons' split: split_if_asm simp del: o_apply) (auto simp: \<pi>_def)
    thus "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (Exists \<phi>)" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by (auto intro!: exI[of _ w] exI[of _ I])
  qed
next
  case (EXISTS \<phi>)
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (EXISTS \<phi>)"
    then obtain w I P where
      *: "x = enc (w, I)" "wf_interp_for_formula (w, I) (EXISTS \<phi>)"
       "length I = n" "length w > 0" "P \<subseteq> {0 .. length w - 1}" "satisfies (w, Inr P # I) \<phi>"
       unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    from *(4,5) have "\<forall>p \<in> P. p < length w" by (cases w) auto
    with *(2-4,6) EXISTS(2) have "enc (w, Inr P # I) \<in> ?L (Suc n) \<phi>"
      by (intro subsetD[OF equalityD1[OF EXISTS(1)], of "Suc n" "enc (w, Inr P # I)"])
       (auto simp: lang\<^isub>M\<^isub>S\<^isub>O_def nth_Cons' split: split_if_asm sum.splits
        intro!: exI[of _ w] exI[of _ "Inr P # I"])
    with *(1) show "x \<in> ?L n (EXISTS \<phi>)"
      by (auto simp: map_index intro!: image_eqI[of _ "map \<pi>"] simp del: o_apply) (auto simp: \<pi>_def)
  next
    fix x assume "x \<in> ?L n (EXISTS \<phi>)"
    then obtain x' where x: "x = map \<pi> x'" and x': "length x' > 0" and "x' \<in> ?L (Suc n) \<phi>" by (auto simp del: o_apply)
    with EXISTS(2) have "x' \<in> lang\<^isub>M\<^isub>S\<^isub>O (Suc n) \<phi>"
      by (intro subsetD[OF equalityD2[OF EXISTS(1)], of "Suc n" x'])
         (auto split: split_if_asm sum.splits)
    then obtain w I' where
      *: "x' = enc (w, I')" "wf_interp_for_formula (w, I') \<phi>" "length I' = Suc n" "satisfies (w, I') \<phi>"
       unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto
    moreover then obtain I\<^isub>0 I where "I' = I\<^isub>0 # I" by (cases I') auto
    moreover with EXISTS(2) *(2) obtain P where "I\<^isub>0 = Inr P"
      by (auto simp: nth_Cons' split: sum.splits split_if_asm)
    moreover have "length w \<ge> 1" using x' *(1) by (cases w) auto
    ultimately have "x = enc (w, I)" "wf_interp_for_formula (w, I) (EXISTS \<phi>)" "length I = n"
      "length w > 0" "satisfies (w, I) (EXISTS \<phi>)" using EXISTS(2) unfolding x
      by (auto simp add: map_tl nth_Cons' split: split_if_asm
        intro!: exI[of _ P] simp del: o_apply) (auto simp: \<pi>_def)
    thus "x \<in> lang\<^isub>M\<^isub>S\<^isub>O n (EXISTS \<phi>)" unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by (auto intro!: exI[of _ w] exI[of _ I])
  qed
qed

lemma wf_rexp_of: "wf_formula n \<phi> \<Longrightarrow> wf n (rexp_of n \<phi>)"
by (induct \<phi> arbitrary: n) (auto simp: wf_rexp_ENC intro: wf_rexp_arbitrary_except split: sum.splits split_if_asm)

lemma wf_rexp_of': "wf_formula n \<phi> \<Longrightarrow> wf n (rexp_of' n \<phi>)"
unfolding rexp_of'_def
by (induct \<phi> arbitrary: n) (auto simp: wf_rexp_ENC intro: wf_rexp_arbitrary_except split: sum.splits split_if_asm)


lemma ENC_Not: "ENC n (Not \<phi>) = ENC n \<phi>"
unfolding ENC_def by auto

lemma ENC_And:
  "wf_formula n (And \<phi> \<psi>) \<Longrightarrow> lang n (ENC n (And \<phi> \<psi>)) - {[]} \<subseteq> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>) - {[]}"
proof
  fix x assume wf: "wf_formula n (And \<phi> \<psi>)" and x: "x \<in> lang n (ENC n (And \<phi> \<psi>)) - {[]}"
  hence wf1: "wf_formula n \<phi>" and wf2: "wf_formula n \<psi>" by auto
  from x obtain w I where wI: "x = enc (w, I)" "wf_interp_for_formula (w, I) (And \<phi> \<psi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  hence "wf_interp_for_formula (w, I) \<phi>" "wf_interp_for_formula (w, I) \<psi>"
    using wf_interp_for_formula_And[OF wf] by auto
  hence "x \<in> (lang n (ENC n \<phi>) - {[]}) \<inter> (lang n (ENC n \<psi>) - {[]})"
    unfolding lang_ENC[OF wf1] lang_ENC[OF wf2] using wI by auto
  thus "x \<in> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>) - {[]}" by blast
qed
  
lemma ENC_Or:
  "wf_formula n (Or \<phi> \<psi>) \<Longrightarrow> lang n (ENC n (Or \<phi> \<psi>)) - {[]} \<subseteq> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>) - {[]}"
proof
  fix x assume wf: "wf_formula n (Or \<phi> \<psi>)" and x: "x \<in> lang n (ENC n (Or \<phi> \<psi>)) - {[]}"
  hence wf1: "wf_formula n \<phi>" and wf2: "wf_formula n \<psi>" by auto
  from x obtain w I where wI: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Or \<phi> \<psi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  hence "wf_interp_for_formula (w, I) \<phi>" "wf_interp_for_formula (w, I) \<psi>"
    using wf_interp_for_formula_Or[OF wf] by auto
  hence "x \<in> (lang n (ENC n \<phi>) - {[]}) \<inter> (lang n (ENC n \<psi>) - {[]})"
    unfolding lang_ENC[OF wf1] lang_ENC[OF wf2] using wI by auto
  thus "x \<in> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>) - {[]}" by blast
qed

lemma project_enc: "map \<pi> (enc (w, x # I)) = enc (w, I)"
unfolding \<pi>_def by auto

lemma ENC_Exists:
  "wf_formula n (Exists \<phi>) \<Longrightarrow> lang n (ENC n (Exists \<phi>)) - {[]} = map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>) - {[]}"
proof (intro equalityI subsetI)
  fix x assume wf: "wf_formula n (Exists \<phi>)" and x: "x \<in> lang n (ENC n (Exists \<phi>)) - {[]}"
  hence wf1: "wf_formula (Suc n) \<phi>" by auto
  from x obtain w I where wI: "x = enc (w, I)" "wf_interp_for_formula (w, I) (Exists \<phi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  with x have "w \<noteq> []" by (cases w) auto
  from wI(2) obtain p where "p < length w" "wf_interp_for_formula (w, Inl p # I) \<phi>"
    using wf_interp_for_formula_Exists[OF wf[folded wI(3)] `w \<noteq> []`] by auto
  with wI(3) have "x \<in> map \<pi> ` (lang (Suc n) (ENC (Suc n) \<phi>) - {[]})"
    unfolding wI(1) lang_ENC[OF wf1] project_enc[symmetric, of w I "Inl p"]
    by (intro imageI CollectI exI[of _ w] exI[of _ "Inl p # I"]) auto
  thus "x \<in> map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>) - {[]}" by blast
next
  fix x assume wf: "wf_formula n (Exists \<phi>)" and "x \<in> map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>) - {[]}"
  hence wf1: "wf_formula (Suc n) \<phi>" and "0 \<in> FOV \<phi>" and x: "x \<in> map \<pi> ` (lang (Suc n) (ENC (Suc n) \<phi>) - {[]})" by auto
  from x obtain w I where wI: "x = map \<pi> (enc (w, I))" "wf_interp_for_formula (w, I) \<phi>" "length I = Suc n"
    using lang_ENC[OF wf1] by auto
  with `0 \<in> FOV \<phi>` obtain p I' where I: "I = Inl p # I'" by (cases I) (fastforce split: sum.splits)+
  with wI have wtlI: "x = enc (w, I')" "length I' = n" unfolding \<pi>_def by auto
  with x have "w \<noteq> []" by (cases w) auto
  have "wf_interp_for_formula (w, I') (Exists \<phi>)"
    using wf_interp_for_formula_Exists[OF wf[folded wtlI(2)] `w \<noteq> []`]
          wf_interp_for_formula_any_Inl[OF wI(2)[unfolded I]] ..
  with wtlI show "x \<in> lang n (ENC n (Exists \<phi>)) - {[]}" unfolding lang_ENC[OF wf] by blast
qed

lemma ENC_EXISTS:
  "wf_formula n (EXISTS \<phi>) \<Longrightarrow> lang n (ENC n (EXISTS \<phi>)) - {[]} = map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>) - {[]}"
proof (intro equalityI subsetI)
  fix x assume wf: "wf_formula n (EXISTS \<phi>)" and x: "x \<in> lang n (ENC n (EXISTS \<phi>)) - {[]}"
  hence wf1: "wf_formula (Suc n) \<phi>" by auto
  from x obtain w I where wI: "x = enc (w, I)" "wf_interp_for_formula (w, I) (EXISTS \<phi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  with x have "w \<noteq> []" by (cases w) auto
  from wI(2) obtain P where "\<forall>p \<in> P. p < length w" "wf_interp_for_formula (w, Inr P # I) \<phi>"
    using wf_interp_for_formula_EXISTS[OF wf[folded wI(3)] `w \<noteq> []`] by auto
  with wI(3) have "x \<in> map \<pi> ` (lang (Suc n) (ENC (Suc n) \<phi>) - {[]})"
    unfolding wI(1) lang_ENC[OF wf1] project_enc[symmetric, of w I "Inr P"]
    by (intro imageI CollectI exI[of _ w] exI[of _ "Inr P # I"]) auto
  thus "x \<in> map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>) - {[]}" by blast
next
  fix x assume wf: "wf_formula n (EXISTS \<phi>)" and "x \<in> map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>) - {[]}"
  hence wf1: "wf_formula (Suc n) \<phi>" and "0 \<in> SOV \<phi>" and x: "x \<in> map \<pi> ` (lang (Suc n) (ENC (Suc n) \<phi>) - {[]})" by auto
  from x obtain w I where wI: "x = map \<pi> (enc (w, I))" "wf_interp_for_formula (w, I) \<phi>" "length I = Suc n"
    using lang_ENC[OF wf1] by auto
  with `0 \<in> SOV \<phi>` obtain P I' where I: "I = Inr P # I'" by (cases I) (fastforce split: sum.splits)+
  with wI have wtlI: "x = enc (w, I')" "length I' = n" unfolding \<pi>_def by auto
  with x have "w \<noteq> []" by (cases w) auto
  have "wf_interp_for_formula (w, I') (EXISTS \<phi>)"
    using wf_interp_for_formula_EXISTS[OF wf[folded wtlI(2)] `w \<noteq> []`]
          wf_interp_for_formula_any_Inr[OF wI(2)[unfolded I]] ..
  with wtlI show "x \<in> lang n (ENC n (EXISTS \<phi>)) - {[]}" unfolding lang_ENC[OF wf] by blast
qed

lemma lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_rexp_of':
  "wf_formula n \<phi> \<Longrightarrow> lang n (rexp_of n \<phi>) - {[]} = lang n (rexp_of' n \<phi>) - {[]}"
unfolding rexp_of'_def proof (induction \<phi> arbitrary: n)
  case (Not \<phi>)
  hence "wf_formula n \<phi>" by simp
  with Not.IH show ?case unfolding rexp_of.simps rexp_of_alt.simps lang.simps ENC_Not by blast
next
  case (And \<phi>\<^isub>1 \<phi>\<^isub>2)
  hence wf1: "wf_formula n \<phi>\<^isub>1" and wf2: "wf_formula n \<phi>\<^isub>2" by force+
  from And.IH(1)[OF wf1] And.IH(2)[OF wf2] show ?case using ENC_And[OF And.prems]
    unfolding rexp_of.simps rexp_of_alt.simps lang.simps rexp_of_list.simps by blast
next
  case (Or \<phi>\<^isub>1 \<phi>\<^isub>2)
  hence wf1: "wf_formula n \<phi>\<^isub>1" and wf2: "wf_formula n \<phi>\<^isub>2" by force+
  from Or.IH(1)[OF wf1] Or.IH(2)[OF wf2] show ?case using ENC_Or[OF Or.prems]
    unfolding rexp_of.simps rexp_of_alt.simps lang.simps by blast
next
  case (Exists \<phi>)
  hence wf: "wf_formula (n + 1) \<phi>" by auto
  have *: "\<And>A. map \<pi> ` A - {[]} = map \<pi> ` (A - {[]})" by auto
  show ?case using ENC_Exists[OF Exists.prems]
    unfolding rexp_of.simps rexp_of_alt.simps lang.simps * Exists.IH[OF wf] by auto
next
  case (EXISTS \<phi>)
  hence wf: "wf_formula (n + 1) \<phi>" by auto
  have *: "\<And>A. map \<pi> ` A - {[]} = map \<pi> ` (A - {[]})" by auto
  show ?case using ENC_EXISTS[OF EXISTS.prems]
    unfolding rexp_of.simps rexp_of_alt.simps lang.simps * EXISTS.IH[OF wf] by auto
qed auto

theorem lang\<^isub>M\<^isub>S\<^isub>O_rexp_of': "wf_formula n \<phi> \<Longrightarrow> lang\<^isub>M\<^isub>S\<^isub>O n \<phi> = lang n (rexp_of' n \<phi>) - {[]}"
unfolding lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_rexp_of'[symmetric] by (rule lang\<^isub>M\<^isub>S\<^isub>O_rexp_of)

end

end
